from __future__ import annotations

import copy
import json
import logging
from ast import literal_eval
from typing import Dict

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_init
from django.utils.translation import gettext_lazy as _
from typing_extensions import TypeAlias

from djblets.db.validators import validate_json
from djblets.util.decorators import cached_property
from djblets.util.serializers import DjbletsJSONEncoder


logger = logging.getLogger(__name__)


#: A type indicating a valid value in JSON data.
#:
#: As of the time this was added, Python does not have its own JSON value
#: definition, and type checkers do not allow for recursive definitions.
#: This is currently very generic, but may be expanded or aliased in the
#: future.
#:
#: Version Added:
#:     3.1
JSONValue: TypeAlias = object


#: A type for a JSON dictionary.
#:
#: Version Added:
#:     3.1
JSONDict: TypeAlias = Dict[str, JSONValue]


class JSONFormField(forms.CharField):
    """Provides a form field for JSON input.

    This is meant to be used by :py:class:`JSONField`, and handles the work of
    rendering JSON content in a more human-editable way for editing (by sorting
    keys and adding indentation), and then validating and returning the
    resulting JSON document back for storage in the field.

    Attributes:
        encoder (json.JSONEncoder):
            The JSON encoder being used to serialize JSON data for editing.
    """

    def __init__(self, encoder=None, encoder_cls=None, encoder_kwargs=None,
                 *args, **kwargs):
        """Initialize the field.

        Args:
            encoder (json.JSONEncoder, optional):
                The explicit JSON encoder instance to use. If specified, this
                takes priority over ``encoder_cls`` and ``encoder_kwargs``,
                and will prevent the field from nicely formatting the
                contents for editing.

                .. deprecated:: 1.0.1

                   ``encoder_cls`` and ``encoder_kwargs`` should be used
                   instead.

            encoder_cls (type, optional):
                The type of encoder to use for serializing the JSON document
                for editing. This must be a subclass of
                :py:class:`json.JSONEncoder`.

            encoder_kwargs (dict, optional):
                Keyword arguments to pass to the constructor for the encoder.

            *args (tuple):
                Extra positional arguments to pass to the field.

            **kwargs (dict):
                Extra keyword arguments to pass to the field.
        """
        super(JSONFormField, self).__init__(*args, **kwargs)

        if encoder is None:
            if encoder_cls is None:
                encoder_cls = DjbletsJSONEncoder
                encoder_kwargs = {
                    'strip_datetime_ms': False,
                }

            encoder = encoder_cls(**dict({
                'indent': 2,
                'separators': (',', ': '),
                'sort_keys': True,
            }, **encoder_kwargs))

        self.encoder = encoder

    def prepare_value(self, value):
        """Prepare a field's value for storage in the database.

        This will encode the value to JSON (unless it's already a string).

        Args:
            value (object):
                The JSON-serializable value to encode.

        Returns:
            unicode: The resulting JSON string.
        """
        if isinstance(value, str):
            return value
        else:
            return self.encoder.encode(value)

    def to_python(self, value):
        """Return the Python representation of the value in the field.

        This will attempt to deserialize the value and return it. If the
        value is not a string, it will be returned directly.

        Args:
            value (object):
                The value stored for the field. This is expected to be a
                string representing serialized JSON data.

        Returns:
            object:
            The deserialized JSON object, or ``None`` if it's an empty string.

        Raises:
            django.core.exceptions.ValidationError:
                The value was not able to be deserialized as JSON content.
        """
        if isinstance(value, str):
            if not value:
                return None

            try:
                return json.loads(value)
            except ValueError as e:
                raise ValidationError(
                    str(e),
                    code='invalid',
                    params={
                        'value': value,
                    })
        else:
            return value


class JSONField(models.TextField):
    """A field for storing JSON-encoded data.

    The data is accessible as standard Python data types and is transparently
    encoded to and decoded from a JSON string in the database.

    Consumers can specify custom encoding behavior by providing an encoder
    class and keyword arguments. By default,
    :py:class:`~djblets.util.serializes.DjbletsJSONEncoder` is used, which
    supports lazy strings, datetimes, and model-specified custom encoding
    behavior.
    """

    serialize_to_string = True
    default_validators = [validate_json]
    default_error_messages = {
        'invalid_type': _('%(type)s is not a supported value type.'),
    }

    def __init__(self, verbose_name=None, name=None, encoder_cls=None,
                 encoder_kwargs=None, **kwargs):
        """Initialize the field.

        Args:
            verbose_name (unicode, optional):
                The verbose name shown for the field.

            name (unicode, optional):
                The attribute name of the field.

            encoder_cls (type, optional):
                The type of encoder to use for serializing the JSON document
                for storage and editing. This must be a subclass of
                :py:class:`json.JSONEncoder`.

            encoder_kwargs (dict, optional):
                Keyword arguments to pass to the constructor for the encoder.
                This may be modified by the field.

            **kwargs (dict):
                Additional keyword arguments for the field.
        """
        super(JSONField, self).__init__(verbose_name,
                                        name,
                                        blank=kwargs.pop('blank', True),
                                        **kwargs)

        if encoder_cls is None:
            encoder_cls = DjbletsJSONEncoder
            encoder_kwargs = {
                'strip_datetime_ms': False,
            }

        self.encoder_cls = encoder_cls
        self.encoder_kwargs = encoder_kwargs or {}
        self.encoder_kwargs['sort_keys'] = True

    @cached_property
    def encoder(self):
        """An encoder instance used for this field."""
        return self.encoder_cls(**self.encoder_kwargs)

    def contribute_to_class(self, cls, name):
        """Add methods/signal handlers to the model owning this field.

        This will add :samp:`get_{fieldname}_json()` and
        :samp:`set_{fieldname}_json()` methods to the model, which will
        allow retrieving and setting the serialized data directly.

        It also listens for when an instance of the model is initialized and
        sets the field to point to the deserialized JSON data.

        Args:
            cls (type):
                The :py:class:`~django.db.models.Model` class to contribute to.

            name (unicode):
                The name of this field on the model.
        """
        def get_json(model_instance):
            return self.value_to_string(model_instance)

        def set_json(model_instance, json):
            setattr(model_instance, self.attname, self.loads(json))

        super(JSONField, self).contribute_to_class(cls, name)

        setattr(cls, 'get_%s_json' % self.name, get_json)
        setattr(cls, 'set_%s_json' % self.name, set_json)

        post_init.connect(self.post_init, sender=cls)

    def pre_save(self, model_instance, add):
        """Return serialized field data to save on a model instance.

        Args:
            model_instance (django.db.models.Model):
                The model instance being saved.

            add (bool, unused):
                Whether this is a new instance being added to the database.
                This is ignored.

        Returns:
            unicode:
            The serialized data to save.
        """
        return self.dumps(getattr(model_instance, self.attname, None))

    def post_init(self, instance, **kwargs):
        """Handle initialization of a model instance.

        This will deserialize the stored data from the database and set it as
        the field data on the model.

        Args:
            instance (django.db.models.Model):
                The model instance being initialized.

            **kwargs (dict):
                Extra keyword arguments from the signal.
        """
        # Check if the field is deferred. If so, we don't want to do anything,
        # as a fetch attempt would result in an SQL query.
        if self.attname not in instance.__dict__:
            return

        value = self.value_from_object(instance)

        if isinstance(value, (dict, list)):
            value = copy.deepcopy(value)
        elif isinstance(value, str):
            value = self.loads(value)
        elif value is None:
            value = {}
        else:
            raise ValidationError(
                self.error_messages['invalid_type'],
                code='invalid',
                params={
                    'type': type(value),
                })

        setattr(instance, self.attname, value)

    def get_prep_value(self, value):
        """Return the serialized value prepared for storage.

        This will serialize the data, if it's a JSON structure and not
        already a string, preparing it for storage in the database or
        another location.

        Args:
            value (object):
                The value to prepare.

        Returns:
            unicode:
            The serialized representation of the value.
        """
        if value is None or isinstance(value, str):
            return value

        return self.dumps(value)

    def value_to_string(self, obj):
        """Return the serialized JSON data from the field.

        Args:
            obj (django.db.models.Model):
                The model instance containing the field.

        Returns:
            unicode:
            The serialized JSON data from the field.
        """
        return self.dumps(self.value_from_object(obj))

    def to_python(self, value):
        """Return a value suitable for using in Python code.

        This will return the deserialized version of the data, allowing
        Python code to work with the data directly.

        Args:
            value (object):
                The value to make usable in Python code.

        Returns:
            object:
            The deserialized data.
        """
        if isinstance(value, str):
            value = self.loads(value)

        return value

    def dumps(self, data):
        """Return serialized version of the provided JSON document.

        If the data is already a string, it will be provided directly.
        Otherwise, it will use the field's associated JSON encoder to serialize
        the data.

        Args:
            data (object):
                The data to serialize.

        Returns:
            unicode:
            The serialized JSON data.
        """
        if isinstance(data, str):
            return data
        else:
            return self.encoder.encode(data)

    def loads(self, val):
        """Return a JSON document from the serialized JSON data.

        This will first attempt to deserialize the JSON data using the standard
        JSON decoder. If it's unable to do so, or it gets back what appears
        to be a double-encoded JSON document or a Python string representation
        of a JSON document, it will attempt to parse the value and return a
        proper representation.

        Args:
            val (unicode):
                The serialized JSON data to deserialize.

        Returns:
            object:
            The deserialized JSON document.
        """
        if not val:
            return {}

        try:
            if not isinstance(val, str):
                val = val.decode(settings.DEFAULT_CHARSET)

            val = json.loads(val)

            # Old versions of JSONField could end up double-encoding JSON
            # data, resulting in a string being stored that then needs to be
            # parsed again. Check for this and try to deserialize. Worst-case,
            # the JSON decoder decides it's actually a string and returns it.
            if isinstance(val, str):
                logger.warning('JSONField decode error. Expected dictionary, '
                               'got string for input "%s"',
                               val)

                val = json.loads(val)
        except ValueError:
            # There's probably embedded unicode markers (like u'foo') in the
            # string, due to bugs in old versions of JSONField. We have to
            # eval it.
            try:
                val = literal_eval(val)
            except Exception as e:
                logger.error('Failed to eval JSONField data "%r": %s',
                             val, e)
                val = {}

            if isinstance(val, str):
                logger.warning('JSONField decode error after literal_eval: '
                               'Expected dictionary, got string: %r',
                               val)
                val = {}

        return val

    def formfield(self, **kwargs):
        """Return a form field that can be used to edit this data.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the subclass.

        Returns:
            JSONFormField:
            The resulting form field.
        """
        formfield_kwargs = dict({
            'form_class': JSONFormField,
            'encoder': None,
            'encoder_cls': self.encoder_cls,
            'encoder_kwargs': self.encoder_kwargs,
        }, **kwargs)

        if self.encoder_cls is None:
            # We're coming from a legacy caller that's providing an encoder
            # instance instead of the class/keyword arguments. Pass that
            # encoder to JSONFormField. It will go into legacy mode and turn
            # off some niceties for editing content.
            formfield_kwargs['encoder'] = self.encoder

        return super(JSONField, self).formfield(**formfield_kwargs)

    def deconstruct(self):
        """Deconstruct the field for Django migrations.

        This makes JSONField migration-safe by encoding the default value
        to a string so that it can be safely loaded into the database.

        .. versionadded:: 0.9

        Returns:
            tuple:
            A tuple of (name, module path, positional arguments, keyword
            arguments).
        """
        name, path, field_args, field_kwargs = \
            super(JSONField, self).deconstruct()

        default = field_kwargs.get('default')

        if default is not None:
            if isinstance(default, type):
                default = self.dumps(default())
            else:
                default = self.dumps(default)

            field_kwargs['default'] = default

        return name, path, field_args, field_kwargs



"""Base class for test cases in Django-based applications."""

import inspect
import os
import re
import socket
import sys
import threading
import warnings
from collections import OrderedDict
from contextlib import contextmanager
from importlib import import_module
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec
from unittest.util import safe_repr

import kgb
from django.apps import apps
from django.conf import settings
from django.core import serializers
from django.core.exceptions import EmptyResultSet, ValidationError
from django.core.handlers.wsgi import WSGIHandler
from django.core.management import call_command
from django.core.servers import basehttp
from django.db import (DatabaseError, DEFAULT_DB_ALIAS, IntegrityError,
                       connections, router)
from django.db.models import Model, Q
from django.db.models.query import MAX_GET_RESULTS
from django.db.models.signals import pre_delete
from django.db.models.sql.compiler import (SQLCompiler,
                                           SQLDeleteCompiler,
                                           SQLInsertCompiler,
                                           SQLUpdateCompiler)
from django.db.models.sql.query import Query as SQLQuery
from django.template import Node
from django.test import testcases

try:
    from django_evolution.models import Evolution, Version
except ImportError:
    Evolution = None
    Version = None

from djblets.siteconfig.models import SiteConfiguration


class StubNodeList(Node):
    def __init__(self, default_text):
        self.default_text = default_text

    def render(self, context):
        return self.default_text


class StubParser:
    def __init__(self, default_text):
        self.default_text = default_text

    def parse(self, until):
        return StubNodeList(self.default_text)

    def delete_first_token(self):
        pass


class TestCase(testcases.TestCase):
    """Base class for test cases.

    Individual tests on this TestCase can use the :py:func:`add_fixtures`
    decorator to add or replace the fixtures used for the test.
    """
    ws_re = re.compile(r'\s+')

    def __call__(self, *args, **kwargs):
        method = getattr(self, self._testMethodName)
        old_fixtures = getattr(self, 'fixtures', None)

        if hasattr(method, '_fixtures'):
            if (getattr(method, '_replace_fixtures') or
                old_fixtures is None):
                self.fixtures = method._fixtures
            else:
                self.fixtures = old_fixtures + method._fixtures

        super(TestCase, self).__call__(*args, **kwargs)

        self.fixtures = old_fixtures

    def shortDescription(self):
        """Returns the description of the current test.

        This changes the default behavior to replace all newlines with spaces,
        allowing a test description to span lines. It should still be kept
        short, though.
        """
        doc = self._testMethodDoc

        if doc is not None:
            doc = doc.split('\n\n', 1)[0]
            doc = self.ws_re.sub(' ', doc).strip()

        return doc

    @contextmanager
    def siteconfig_settings(self, settings):
        """Temporarily sets siteconfig settings for a test.

        Subclasses should override this if they want to run a method like
        :py:func:`~djblets.siteconfig.django_settings.apply_django_settings`
        before and after each test.

        Args:
            settings (dict):
                The new siteconfig settings to set.

        Context:
            The current site configuration will contain the new settings for
            this test.
        """
        siteconfig = SiteConfiguration.objects.get_current()

        old_settings = {}

        for key, value in settings.items():
            old_settings[key] = siteconfig.get(key)
            siteconfig.set(key, value)

        siteconfig.save()

        try:
            yield
        finally:
            for key, value in old_settings.items():
                siteconfig.set(key, value)

            siteconfig.save()

    def assertAttrsEqual(self, obj, attrs, msg=None):
        """Assert that attributes on an object match expected values.

        This will compare each attribute defined in ``attrs`` against the
        corresponding attribute on ``obj``. If the attribute value does not
        match, or the attribute is missing, this will assert.

        Args:
            obj (object):
                The object to compare attribute values to.

            attrs (dict):
                A dictionary of expected attribute names and values.

            msg (unicode, optional):
                A custom message to include if there's a failure.

        Raises:
            AssertionError:
                An attribute was not found or the value did not match.
        """
        for key, value in attrs.items():
            try:
                attr_value = getattr(obj, key)

                if attr_value != value:
                    raise self.failureException(self._formatMessage(
                        msg,
                        '%s: %s != %s'
                        % (key, safe_repr(attr_value),
                           safe_repr(value))))
            except AttributeError:
                raise self.failureException(self._formatMessage(
                    msg,
                    'Attribute "%s" was not found on %s'
                    % (key, safe_repr(obj))))

    def assertRaisesValidationError(self, expected_messages, *args, **kwargs):
        """Assert that a ValidationError is raised with the given message(s).

        This is a wrapper around :py:meth:`assertRaisesMessage` with a
        :py:class:`ValidationError` that handles converting the expected
        messages into a list (if it isn't already) and then converting that
        into a string representation, which is what
        :py:meth:`assertRaisesMessage` will be checking against.

        Args:
            expected_messages (list or unicode):
                The expected messages as either a list of strings or a
                single string.

            args:
                Additional arguments to pass to :py:meth:`assertRaisesMessage`.

            kwargs:
                Additional keyword arguments to pass to
                :py:meth:`assertRaisesMessage`.
        """
        if isinstance(expected_messages, str):
            expected_messages = [expected_messages]

        return self.assertRaisesMessage(ValidationError,
                                        repr(expected_messages),
                                        *args,
                                        **kwargs)

    def assertRaisesMessage(self, expected_exception, expected_message,
                            *args, **kwargs):
        """Assert that an exception is raised with a given message.

        This is a replacement for Django's assertRaisesMessage that behaves
        well with a design change in Python 2.7.9/10, without crashing.
        """
        # The difference between this version and Django's is that we're
        # not taking the callable_obj as an argument with a default value and
        # passing it down to assertRaisesRegex. Python 2.7.9/10's
        # implementation defaults callable_obj to a special value, which
        # Django overrides.
        return self.assertRaisesRegex(expected_exception,
                                      re.escape(expected_message),
                                      *args, **kwargs)

    @contextmanager
    def assertWarns(self, cls=DeprecationWarning, message=None):
        """Assert that a warning is generated with a given message.

        This method only supports code which generates a single warning.
        Tests which make use of code generating multiple warnings will
        need to manually catch their warnings.

        Args:
            cls (type, optional):
                The expected warning type.

            message (unicode, optional):
                The expected error message, if any.

        Context:
            The test to run.
        """
        with warnings.catch_warnings(record=True) as w:
            # Some warnings such as DeprecationWarning are filtered by
            # default, stop filtering them.
            warnings.simplefilter('always')

            # We do need to ignore this one, though, or a lot of things will
            # fail. This is specific to Python 3.10 and the versions of six
            # in Django.
            warnings.filterwarnings(
                'ignore',
                r'_SixMetaPathImporter.find_spec\(\) not found')

            self.assertEqual(len(w), 0)

            yield

            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, cls))

            if message is not None:
                self.assertEqual(message, str(w[-1].message))

    @contextmanager
    def assertNoWarnings(self):
        """Assert that a warning is not generated.

        Context:
            The test to run.
        """
        with warnings.catch_warnings(record=True) as w:
            # Some warnings such as DeprecationWarning are filtered by
            # default, stop filtering them.
            warnings.simplefilter('always')

            # We do need to ignore this one, though, or a lot of things will
            # fail. This is specific to Python 3.10 and the versions of six
            # in Django.
            warnings.filterwarnings(
                'ignore',
                r'_SixMetaPathImporter.find_spec\(\) not found')

            self.assertEqual(len(w), 0)

            yield

            self.assertEqual(len(w), 0)

    @contextmanager
    def assertQueries(self, queries, num_statements=None):
        """Assert the number and complexity of queries.

        This provides advanced checking of queries, allowing the caller to
        match filtering, JOINs, ordering, selected fields, and more.

        This takes a list of dictionaries with query information. Each
        contains the following:

        Keys:
            model (type):
                The model representing the results that would be returned
                or altered by the query.

            annotations (dict, optional):
                A dictionary containing applied annotations.

                Keys are destination attribute names, and values are the
                annotation instances.

                The default is empty.

            distinct (bool, optional):
                Whether :py:meth:`django.db.models.query.QuerySet.distinct`
                was used.

                The default is ``False``.

            distinct_fields (tuple of str, optional):
                A list of fields passed to
                :py:meth:`django.db.models.query.QuerySet.distinct`.

                The default is empty.

            extra (dict, optional):
                State passed in calls to
                :py:meth:`django.db.models.query.QuerySet.extra` when using
                ``select`` and ``select_params``.

                Each key maps to a key in ``select``, and each value is a
                tuple containing the value in ``select`` and the corresponding
                value (if any) in ``select_params``.

                Values are normalized to collapse and strip whitespace, to
                help with comparison.

                The default is empty.

            extra_order_by (list of str, optional):
                State passed in calls to
                :py:meth:`django.db.models.query.QuerySet.extra` when using
                ``order_by``.

                The default is empty.

            extra_tables (list of str, optional):
                State passed in calls to
                :py:meth:`django.db.models.query.QuerySet.extra` when using
                ``tables``.

                The default is empty.

            group_by (bool or tuple, optional):
                Whether no fields will be grouped (``None``), all fields will
                be grouped (``True``), or specific expressions/field names are
                grouped (a tuple).

                This is influenced by using
                :py:meth:`django.db.models.query.QuerySet.annotate`.

                The default is ``None``.

            limit (int, optional):
                The value for a ``LIMIT`` in the ``SELECT``.

                This will generally only need to be supplied if testing a
                query using :py:meth:`QuerySet.exists()
                <django.db.models.query.QuerySet.exists>` or when slicing
                results.

                Django itself sometimes uses a default of ``None`` and
                sometimes a default currently of ``21`` (this exact value,
                and when it's used, is considered an implementation detail
                in Django). Both of these will match a caller-provided
                ``limit`` value of ``None``.

                The default is ``None``.

            num_joins (int, optional):
                The number of tables JOINed.

                The default is 0.

            offset (int, optional):
                The value for an ``OFFSET`` in the ``SELECT``.

                The default is 0.

            only_fields (set of str, optional);
                The specific fields being fetched, or ``None`` if fetching
                all fields.

                The default is ``None``.

            order_by (tuple of str, optional):
                The ordering criteria.

                The default is empty.

            select_for_update (bool, optional):
                Whether this is a select-for-update operation.

                The default is ``False``.

            select_related (set of str, optional):
                The table names involved in a
                :py:meth:`django.db.models.query.QuerySet.select_related`.

            subquery (bool, optional):
                Whether this is considered a subquery of another query.

                The default is ``False``.

            tables (set of str, optional):
                The tables involved in the query.

                The default is the model's table name.

            type (str, optional):
                The query type. This would be one of ``DELETE``, ``INSERT``,
                ``SELECT``, or ``UPDATE``.

                The default is ``SELECT``.

            values_select (list of str, optional):
                A list of specified fields being returned using
                :py:meth:`~django.db.models.query.QuerySet.values` or
                :py:meth:`~django.db.models.query.QuerySet.values_list` or

            where (django.db.models.Q, optional):
                The query expression objects used to represent the filter on
                the query.

        Version Added:
            3.0

        Args:
            queries (list of dict):
                The list of query dictionaries to compare executed queries
                against.

            num_statements (int, optional):
                The numbre of SQL statements executed.

                This defaults to the length of ``queries``, but callers may
                need to provide an explicit number, as some operations may add
                additional database-specific statements (such as
                transaction-related SQL) that won't be covered in ``queries``.

        Raises:
            AssertionError:
                The parameters passed, or the queries compared, failed
                expectations.
        """
        assert isinstance(queries, list)

        if num_statements is None:
            num_statements = len(queries)

        spy_agency = kgb.SpyAgency()

        executed_queries = []
        queries_to_qs = {}

        # Track Query objects any time a compiler is executing SQL.
        @spy_agency.spy_for(SQLCompiler.execute_sql,
                            owner=SQLCompiler)
        def _sql_compiler_execute_sql(_self, *args, **kwargs):
            if isinstance(_self, SQLDeleteCompiler):
                query_type = 'DELETE'
            elif isinstance(_self, SQLUpdateCompiler):
                query_type = 'UPDATE'
            else:
                query_type = 'SELECT'

            executed_queries.append((query_type, _self.query))

            return SQLCompiler.execute_sql.call_original(
                _self, *args, **kwargs)

        @spy_agency.spy_for(SQLInsertCompiler.execute_sql,
                            owner=SQLInsertCompiler)
        def _sql_insert_compiler_execute_sql(_self, *args, **kwargs):
            executed_queries.append(('INSERT', _self.query))

            return SQLInsertCompiler.execute_sql.call_original(
                _self, *args, **kwargs)

        # Build and track Q() objects any time they're added to a Query.
        @spy_agency.spy_for(SQLQuery.add_q, owner=SQLQuery)
        def _query_add_q(_self, q_object):
            try:
                queries_to_qs[_self] &= q_object
            except KeyError:
                queries_to_qs[_self] = q_object

            return SQLQuery.add_q.call_original(_self, q_object)

        # Copy Q() objects any time a Query is cloned.
        @spy_agency.spy_for(SQLQuery.clone, owner=SQLQuery)
        def _query_clone(_self, *args, **kwargs):
            result = SQLQuery.clone.call_original(_self, *args, **kwargs)

            try:
                queries_to_qs[result] = queries_to_qs[_self]
            except KeyError:
                pass

            return result

        # Set up an explicit pre_delete signal. During deletion, Django
        # attempts to determine if it can fast-delete (which can be done if,
        # amongst other things, signals don't need to be emitted for each
        # object).
        #
        # This can lead to inconsistencies in test runs, if registration of
        # pre_delete or post_delete signals is conditional on that run. We
        # want to ensure a stable query count and information for DELETEs,
        # so we deny fast-deletion by setting up a model-global signal handler
        # during capture of queries.
        def _on_pre_delete(**kwargs):
            pass

        pre_delete.connect(_on_pre_delete)

        # Check that the number of SQL queries matches expectations.
        try:
            with self.assertNumQueries(num_statements):
                yield
        finally:
            # We no longer need to track anything in the compiler of Query.
            spy_agency.unspy_all()
            pre_delete.disconnect(_on_pre_delete)

        # Make sure we received the expected number of Queries.
        #
        # First we have to convert each to SQL and see what we get. Any
        # with EmptyResultSet will be skipped.
        query_sqls = []
        norm_executed_queries = []

        for executed_query_info in executed_queries:
            executed_query = executed_query_info[1]

            # First thing we want to do is grab the SQL. This may fail, and
            # if it does, it represents a query that we've caught that isn't
            # actually going to be executed (likely a component of another).
            try:
                try:
                    query_sqls.append([str(executed_query)])
                except ValueError:
                    # When doing an INSERT OR IGNORE (SQLite), the SQL
                    # can be a list rather than a string. Query.__str__
                    # doesn't know how to handle this, and will crash.
                    # We'll deal with it ourselves.
                    sql_statements = executed_query.sql_with_params()
                    assert isinstance(sql_statements, list)

                    query_sqls.append([
                        _sql % _sql_params
                        for _sql, _sql_params in sql_statements
                    ])

                norm_executed_queries.append(executed_query_info)
            except EmptyResultSet:
                # This will be skipped.
                pass

        executed_queries = norm_executed_queries

        # Now we can compare numbers.
        error_lines = []
        num_queries = len(queries)
        num_executed_queries = len(executed_queries)

        if num_queries != num_executed_queries:
            error_lines += [
                '%s queries were provided, but %s were executed.'
                % (num_queries, num_executed_queries),

                '',
            ]

            if num_queries < num_executed_queries:
                queries += [
                    {}
                    for _i in range(num_executed_queries - num_queries)
                ]
            elif num_queries > num_executed_queries:
                queries = queries[:num_executed_queries]

        # Go through each matching Query and compare state.
        ws_re = self.ws_re
        all_failures = []
        queries_iter = enumerate(zip(queries, executed_queries, query_sqls))

        for (i,
             (query_info,
              (executed_query_type, executed_query),
              query_sql)) in queries_iter:
            # Check if this is set to be skipped, as per the above checks.
            if query_sql is None:
                continue

            failures = []

            executed_table_name = executed_query.model._meta.db_table
            executed_reffed_tables = {
                _table_name
                for _table_name in executed_query.alias_map.keys()
                if executed_query.alias_refcount.get(_table_name, 0) > 0
            }

            # Check 'model'.
            model = query_info.get('model', None)

            if model is not executed_query.model:
                failures.append(('model', model, executed_query.model))

            # Check the query type.
            query_type = query_info.get('type', 'SELECT')

            if query_type != executed_query_type:
                failures.append(('type', query_type, executed_query_type))

            # Check the fields that are easy to compare.
            for key, default in (('annotations', {}),
                                 ('distinct', False),
                                 ('distinct_fields', ()),
                                 ('extra_order_by', ()),
                                 ('extra_tables', ()),
                                 ('group_by', None),
                                 ('order_by', ()),
                                 ('select_for_update', False),
                                 ('subquery', False),
                                 ('values_select', ())):
                value = query_info.get(key, default)
                executed_value = getattr(executed_query, key)

                if value != executed_value:
                    failures.append((key, value, executed_value))

            # Check 'extra'.
            value = {
                _key: (ws_re.sub(' ', _value[0]).strip(), _value[1])
                for _key, _value in query_info.get('extra', {}).items()
            }
            executed_value = {
                _key: (ws_re.sub(' ', _value[0]).strip(), _value[1])
                for _key, _value in executed_query.extra.items()
            }

            if value != executed_value:
                failures.append(('extra', value, executed_value))

            # Check 'offset'.
            value = query_info.get('offset', 0)
            executed_value = executed_query.low_mark

            if value != executed_value:
                failures.append(('offset', value, executed_value))

            # Check 'limit'.
            value = query_info.get('limit', None)
            executed_value = executed_query.high_mark

            if (value != executed_value and
                (value is not None or executed_value != MAX_GET_RESULTS)):
                failures.append(('limit', value, executed_value))

            # Check 'num_joins'.
            value = query_info.get('num_joins', 0)

            if executed_query.alias_map:
                # Consider all tables that have been JOINed, ignoring the
                # main table.
                executed_value = len(executed_reffed_tables -
                                     {executed_table_name})
            else:
                executed_value = 0

            if value != executed_value:
                failures.append(('num_joins', value, executed_value))

            # Check 'tables'.
            if model is None:
                default = {}
            else:
                default = {model._meta.db_table}

            value = set(query_info.get('tables', default))

            if executed_query.alias_refcount:
                executed_value = executed_reffed_tables | {executed_table_name}
            else:
                executed_value = {executed_table_name}

            if value != executed_value:
                failures.append(('tables', value, executed_value))

            # Check 'only_fields'.
            value = set(query_info.get('only_fields', set()))
            executed_value = set(executed_query.deferred_loading[0])

            if value != executed_value:
                failures.append(('only_fields', value, executed_value))

            # Check 'select_related'.
            value = set(query_info.get('select_related', set()))
            executed_value = executed_query.select_related

            if executed_value:
                executed_value = set(executed_value.keys())
            else:
                executed_value = set()

            if value != executed_value:
                failures.append(('select_related', value, executed_value))

            # Check 'where'. the Q object used to filter.
            value = query_info.get('where', Q())
            executed_value = queries_to_qs.get(executed_query, Q())

            if (executed_value and
                len(executed_value.children) == 1 and
                isinstance(executed_value.children[0], Q)):
                # filter() created a Q() containing nothing but a nested Q().
                # This is annoying to compare, and could always change in the
                # future. Extract the inner Q() and use that for comparison
                # purposes.
                executed_value = executed_value.children[0]

            if value != executed_value:
                failures.append(('where', value, executed_value))

            if failures:
                all_failures.append((i, executed_query, failures, query_sql))

        # Check if we found any failures, and include them in an assertion.
        if all_failures:
            num_failures = len(all_failures)

            if num_failures == 1:
                error_lines.append('1 query failed to meet expectations.')
            else:
                error_lines.append(
                    '%s queries failed to meet expectations.'
                    % num_failures,
                )

            for i, executed_query, failures, query_sql in all_failures:
                if failures:
                    error_lines += [
                        '',
                        'Query %s:' % (i + 1),
                    ] + [
                        '  %s: %s != %s' % _failure
                        for _failure in failures
                    ] + [
                        '  SQL: %s' % _sql
                        for _sql in query_sql
                    ]

            self.fail('\n'.join(error_lines))


class TestModelsLoaderMixin(object):
    """Allows unit test modules to provide models to test against.

    This allows a unit test file to provide models that will be synced to the
    database and flushed after tests. These can be tested against in any unit
    tests.

    Typically, Django requires any test directories to be pre-added to
    INSTALLED_APPS in order for models to be created in the test database.

    This mixin works around this by dynamically adding the module to
    INSTALLED_APPS and forcing the database to be synced. It also will
    generate a fake 'models' module to satisfy Django's requirement, if one
    doesn't already exist.

    By default, this will assume that the test class's module is the one that
    should be added to INSTALLED_APPS. This can be changed by overriding
    :py:attr:`tests_app`.
    """
    tests_app = None

    @classmethod
    def setUpClass(cls):
        cls._tests_loader_models_mod = None

        if not cls.tests_app:
            cls.tests_app = cls.__module__

        models_mod_name = '%s.models' % cls.tests_app

        try:
            models_mod = import_module(models_mod_name)
        except ImportError:
            # Set up a 'models' module, containing any models local to the
            # module that this TestCase is in.

            # It's not enough to simply create a module type. We need to
            # create a basic spec, and then we need to have the module
            # system create a module from it. There's a handy public
            # function to do this on Python 3.5, but Python 3.4 lacks a
            # public function. Fortunately, it's easy to call a private
            # one.
            spec = ModuleSpec(name=models_mod_name,
                              loader=None)

            models_mod = module_from_spec(spec)
            assert models_mod

            # Transfer all the models over into this new module.
            module_name = cls.__module__
            test_module = sys.modules[module_name]

            for key, value in test_module.__dict__.items():
                if (inspect.isclass(value) and
                    issubclass(value, Model) and
                    value.__module__ == module_name):
                    models_mod.__dict__[key] = value

        cls._tests_loader_models_mod = models_mod

        if models_mod:
            sys.modules[models_mod.__name__] = models_mod

        cls._models_loader_old_settings = settings.INSTALLED_APPS
        settings.INSTALLED_APPS = list(settings.INSTALLED_APPS) + [
            cls.tests_app,
        ]

        # If Django Evolution is being used, we'll want to clear out any
        # recorded schema information so that it can be generated from
        # scratch when we set up the database.
        if (Evolution is not None and
            Version is not None and
            getattr(settings, 'DJANGO_EVOLUTION_ENABLED', True)):
            Version.objects.all().delete()
            Evolution.objects.all().delete()

        # Push the new set of installed apps, and begin registering
        # each of the models associated with the tests.
        apps.set_installed_apps(settings.INSTALLED_APPS)
        app_config = apps.get_containing_app_config(cls.tests_app)

        if models_mod:
            app_label = app_config.label

            for key, value in models_mod.__dict__.items():
                if inspect.isclass(value) and issubclass(value, Model):
                    # The model was likely registered under another app,
                    # so we need to remove the old one and add the new
                    # one.
                    try:
                        del apps.all_models[value._meta.app_label][
                            value._meta.model_name]
                    except KeyError:
                        pass

                    value._meta.app_label = app_label
                    apps.register_model(app_label, value)

        call_command('migrate', run_syncdb=True, verbosity=0,
                     interactive=False)

        super(TestModelsLoaderMixin, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        super(TestModelsLoaderMixin, cls).tearDownClass()

        call_command('flush', verbosity=0, interactive=False)

        settings.INSTALLED_APPS = cls._models_loader_old_settings

        # If we added a fake 'models' module to sys.modules, remove it.
        models_mod = cls._tests_loader_models_mod

        if models_mod:
            try:
                del sys.modules[models_mod.__name__]
            except KeyError:
                pass

        apps.unset_installed_apps()
        apps.all_models[cls.tests_app].clear()

        # Set this free so the garbage collector can eat it.
        cls._tests_loader_models_mod = None


class FixturesCompilerMixin(object):
    """Compiles and efficiently loads fixtures into a test suite.

    Unlike Django's standard fixture support, this doesn't re-discover
    and re-deserialize the referenced fixtures every time they're needed.
    Instead, it precompiles the fixtures the first time they're found and
    reuses their objects for future tests.

    However, also unlike Django's, this does not accept compressed or
    non-JSON fixtures.
    """

    _precompiled_fixtures = {}
    _fixture_dirs = []

    def _fixture_setup(self):
        """Set up fixtures for unit tests."""
        # Temporarily hide the fixtures, so that the parent class won't
        # do anything with them.
        self._hide_fixtures = True
        super(FixturesCompilerMixin, self)._fixture_setup()
        self._hide_fixtures = False

        if getattr(self, 'multi_db', False):
            databases = connections
        else:
            databases = [DEFAULT_DB_ALIAS]

        if hasattr(self, 'fixtures'):
            for db in databases:
                self.load_fixtures(self.fixtures, db=db)

    def load_fixtures(self, fixtures, db=DEFAULT_DB_ALIAS):
        """Load fixtures for the current test.

        This is called for every fixture in the test case's ``fixtures``
        list. It can also be called by an individual test to add additional
        fixtures on top of that.

        Args:
            fixtures (list of unicode):
                The list of fixtures to load.

            db (unicode):
                The database name to load fixture data on.
        """
        if not fixtures:
            return

        if db not in self._precompiled_fixtures:
            self._precompiled_fixtures[db] = {}

        for fixture in fixtures:
            if fixture not in self._precompiled_fixtures[db]:
                self._precompile_fixture(fixture, db)

        self._load_fixtures(fixtures, db)

    def _precompile_fixture(self, fixture, db):
        """Precompile a fixture.

        The fixture is loaded and deserialized, and the resulting objects
        are stored for future use.

        Args:
            fixture (unicode):
                The name of the fixture.

            db (unicode):
                The database name to load fixture data on.
        """
        assert db in self._precompiled_fixtures
        assert fixture not in self._precompiled_fixtures[db]

        fixture_path = None

        for fixture_dir in self._get_fixture_dirs():
            fixture_path = os.path.join(fixture_dir, fixture + '.json')

            if os.path.exists(fixture_path):
                break

        try:
            if not fixture_path:
                raise IOError('Fixture path not found')

            with open(fixture_path, 'r') as fp:
                self._precompiled_fixtures[db][fixture] = [
                    obj
                    for obj in serializers.deserialize('json', fp, using=db)
                    if ((hasattr(router, 'allow_syncdb') and
                         router.allow_syncdb(db, obj.object.__class__)) or
                        (hasattr(router, 'allow_migrate_model') and
                         router.allow_migrate_model(db, obj.object)))
                ]
        except IOError as e:
            sys.stderr.write('Unable to load fixture %s: %s\n' % (fixture, e))

    def _get_fixture_dirs(self):
        """Return the list of fixture directories.

        This is computed only once and cached.

        Returns:
            The list of fixture directories.
        """
        if not self._fixture_dirs:
            app_module_paths = []

            for app in self._get_app_model_modules():
                if hasattr(app, '__path__'):
                    # It's a 'models/' subpackage.
                    for path in app.__path__:
                        app_module_paths.append(path)
                else:
                    # It's a models.py module
                    app_module_paths.append(app.__file__)

            all_fixture_dirs = [
                os.path.join(os.path.dirname(path), 'fixtures')
                for path in app_module_paths
            ]

            self._fixture_dirs = [
                fixture_dir
                for fixture_dir in all_fixture_dirs
                if os.path.exists(fixture_dir)
            ]

        return self._fixture_dirs

    def _load_fixtures(self, fixtures, db):
        """Load precompiled fixtures.

        Each precompiled fixture is loaded and then used to populate the
        database.

        Args:
            fixtures (list of unicode):
                The list of fixtures to load.

            db (unicode):
                The database name to load fixture data on.
        """
        table_names = set()
        connection = connections[db]

        with connection.constraint_checks_disabled():
            for fixture in fixtures:
                assert db in self._precompiled_fixtures
                assert fixture in self._precompiled_fixtures[db]
                deserialized_objs = self._precompiled_fixtures[db][fixture]

                to_save = OrderedDict()
                to_save_m2m = []

                # Start off by going through the list of deserialized object
                # information from the fixtures and group them into categories
                # that we'll further filter down later.
                for deserialized_obj in deserialized_objs:
                    obj = deserialized_obj.object

                    table_names.add(obj._meta.db_table)
                    to_save.setdefault(type(obj), []).append(obj)

                    if deserialized_obj.m2m_data:
                        to_save_m2m.append((obj, deserialized_obj.m2m_data))

                # Now we'll break this down into objects we can batch-create
                # and ones we have to update.
                #
                # The database may already have entries for some of these
                # models, particularly if they're being shared across multiple
                # tests for the same suite. In that case, when doing a save()
                # on an object, Django will perform a update-or-create
                # operation, involving an UPDATE statement followed by an
                # INSERT If the UPDATE returned 0 rows.
                #
                # That's wasteful. Instead, we're going to just delete
                # everything matching the deserialized objects' IDs and then
                # bulk-create new entries.
                try:
                    for model_cls, objs in to_save.items():
                        obj_ids = [
                            _obj.pk
                            for _obj in objs
                            if _obj.pk
                        ]

                        if obj_ids:
                            model_cls.objects.filter(pk__in=obj_ids).delete()

                        model_cls.objects.using(db).bulk_create(objs)
                except (DatabaseError, IntegrityError) as e:
                    meta = model_cls._meta
                    sys.stderr.write(
                        'Could not load %s.%s from fixture "%s": %s\n'
                        % (meta.app_label,
                           meta.object_name,
                           fixture,
                           e))
                    raise

                # Now save any Many-to-Many field relations.
                #
                # Note that we're not assigning the values to the
                # ManyToManyField attribute. Instead, we're getting the add()
                # method and calling that. We can trust that the relations are
                # empty (since we've newly-created the object), so we don't
                # need to clear the old list first.
                for obj, m2m_data in to_save_m2m:
                    try:
                        for m2m_attr, m2m_objs in m2m_data.items():
                            getattr(obj, m2m_attr).add(*m2m_objs)
                    except (DatabaseError, IntegrityError) as e:
                        meta = obj._meta
                        sys.stderr.write(
                            'Could not load Many-to-Many entries for %s.%s '
                            '(pk=%s) in fixture "%s": %s\n'
                            % (meta.app_label,
                               meta.object_name,
                               obj.pk,
                               fixture,
                               e))
                        raise

        # We disabled constraints above, so check now.
        connection.check_constraints(table_names=sorted(table_names))

    def _get_app_model_modules(self):
        """Return the entire list of registered Django app models modules.

        This is an internal utility function that's used to provide
        compatibility with all supported versions of Django.

        Returns:
            list:
            The list of Django applications.
        """
        return [
            app.models_module
            for app in apps.get_app_configs()
            if app.models_module
        ]

    def __getattribute__(self, name):
        if name == 'fixtures' and self.__dict__.get('_hide_fixtures'):
            raise AttributeError

        return super(FixturesCompilerMixin, self).__getattribute__(name)


class TagTest(TestCase):
    """Base testing setup for custom template tags"""

    def setUp(self):
        self.parser = StubParser(self.getContentText())

    def getContentText(self):
        return "content"


# The following is all based on the code at
# http://trac.getwindmill.com/browser/trunk/windmill/authoring/djangotest.py,
# which is based on the changes submitted for Django in ticket 2879
# (http://code.djangoproject.com/ticket/2879)
#
# A lot of this can go away when/if this patch is committed to Django.

# Code from django_live_server_r8458.diff
#     @ http://code.djangoproject.com/ticket/2879#comment:41
# Editing to monkey patch django rather than be in trunk

class StoppableWSGIServer(basehttp.WSGIServer):
    """
    WSGIServer with short timeout, so that server thread can stop this server.
    """
    def server_bind(self):
        """Sets timeout to 1 second."""
        basehttp.WSGIServer.server_bind(self)
        self.socket.settimeout(1)

    def get_request(self):
        """Checks for timeout when getting request."""
        try:
            sock, address = self.socket.accept()
            sock.settimeout(None)
            return (sock, address)
        except socket.timeout:
            raise


class WSGIRequestHandler(basehttp.WSGIRequestHandler):
    """A custom WSGIRequestHandler that logs all output to stdout.

    Normally, WSGIRequestHandler will color-code messages and log them
    to stderr. It also filters out admin and favicon.ico requests. We don't
    need any of this, and certainly don't want it in stderr, as we'd like
    to only show it on failure.
    """
    def log_message(self, format, *args):
        print(format % args)


class TestServerThread(threading.Thread):
    """Thread for running a http server while tests are running."""

    def __init__(self, address, port):
        self.address = address
        self.port = port
        self._stopevent = threading.Event()
        self.started = threading.Event()
        self.error = None
        super(TestServerThread, self).__init__()

    def run(self):
        """
        Sets up test server and database and loops over handling http requests.
        """
        try:
            handler = basehttp.AdminMediaHandler(WSGIHandler())
            server_address = (self.address, self.port)
            httpd = StoppableWSGIServer(server_address,
                                        WSGIRequestHandler)
            httpd.set_app(handler)
            self.started.set()
        except basehttp.WSGIServerException as e:
            self.error = e
            self.started.set()
            return

        # Must do database stuff in this new thread if database in memory.
        from django.conf import settings

        if hasattr(settings, 'DATABASES'):
            db_engine = settings.DATABASES['default']['ENGINE']
            test_db_name = settings.DATABASES['default']['TEST_NAME']
        else:
            db_engine = settings.DATABASE_ENGINE
            test_db_name = settings.TEST_DATABASE_NAME

        if (db_engine.endswith('sqlite3') and
            (not test_db_name or test_db_name == ':memory:')):
            # Import the fixture data into the test database.
            if hasattr(self, 'fixtures'):
                # We have to use this slightly awkward syntax due to the fact
                # that we're using *args and **kwargs together.
                testcases.call_command('loaddata', verbosity=0, *self.fixtures)

        # Loop until we get a stop event.
        while not self._stopevent.isSet():
            httpd.handle_request()

    def join(self, timeout=None):
        """Stop the thread and wait for it to finish."""
        self._stopevent.set()
        threading.Thread.join(self, timeout)



"""Managers for API-related database models."""

import json
import logging

from django.db import IntegrityError
from django.db.models import Manager, Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from djblets.deprecation import RemovedInDjblets40Warning
from djblets.secrets.token_generators import token_generator_registry
from djblets.secrets.token_generators.legacy_sha1 import \
    LegacySHA1TokenGenerator
from djblets.webapi.errors import WebAPITokenGenerationError
from djblets.webapi.signals import webapi_token_created


logger = logging.getLogger(__name__)


class WebAPITokenManager(Manager):
    """Manages WebAPIToken models."""

    def generate_token(self,
                       user,
                       max_attempts=20,
                       note=None,
                       policy={},
                       auto_generated=False,
                       expires=None,
                       token_generator_id=None,
                       token_info=None,
                       **kwargs):
        """Generate a WebAPIToken for a user.

        This will attempt to construct a unique WebAPIToken for a user.

        Since a collision is possible, it will try up to a certain number
        of times. If it cannot create a unique token, a
        :py:class:`~djblets.webapi.errors.WebAPITokenGenerationError` will be
        raised.

        Version Changed:
            3.0:
            * Added the ``token_generator_id`` and ``token_info`` parameters.
              These are used to specify what type of token to generate. These
              parameters are currently optional but will be required in
              Djblets 4.0.

            * Added the optional ``expires`` parameter. This can be used to
              set an expiration date for the token.

        Args:
            user (django.contrib.auth.models.User):
                The user who will own the token.

            max_attempts (int, optional):
                The maximum number of attempts to try to find a non-conflicting
                token. Defaults to 20.

            note (unicode, optional):
                A note describing the token.

            policy (dict, optional):
                The policy document describing what this token can access
                in the API. By default, this provides full access.

            auto_generated (bool, optional):
                Whether or not the token is being automatically generated.

                This parameter does not affect token generation. It's value
                will solely be used in the
                :py:data:`~djblets.webapi.signals.webapi_token_created` signal.

            expires (datetime.datetime, optional):
                The date and time that the token will expire. By default the
                token will never expire.

                Version Added:
                    3.0

            token_generator_id (str, optional):
                The ID of the token generator to use for generating the token.
                If not set, this will default to :py:class:
                `djblets.secrets.token_generators.legacy_sha1
                .LegacySHA1TokenGenerator`'s ID.

                This parameter will be required in Djblets 4.0.

                Version Added:
                    3.0

            token_info (dict, optional):
                A dictionary that contains information needed for token
                generation. If not set, this will default to a dictionary that
                contains necessary information for the :py:class:
                `djblets.secrets.token_generators.legacy_sha1
                .LegacySHA1TokenGenerator`.

                See :py:mod:`djblets.secrets.token_generators` for further
                details on what is required here for each token generator.

                This parameter will be required in Djblets 4.0.

                Version Added:
                    3.0

            **kwargs (dict):
                Additional keyword arguments representing fields in the token.
                These will be set on the token object.

        Returns:
            djblets.webapi.models.BaseWebAPIToken:
            The generated API token.

        Raises:
            djblets.webapi.errors.WebAPITokenGenerationError:
                The token was not able to be generated after ``max_attempts``
                number of collisions were hit.

            KeyError:
                The ``token_info`` dictionary is missing a required key.
        """
        if token_generator_id is None:
            token_generator_id = LegacySHA1TokenGenerator.token_generator_id

            RemovedInDjblets40Warning.warn(
                'The token_generator_id parameter was not set. This '
                'parameter will be required in Djblets 4.0.')

        if token_info is None:
            token_info = {
                'user': user,
                'attempt': 0,
            }

            RemovedInDjblets40Warning.warn(
                'The token_info parameter was not set. This parameter '
                'will be required in Djblets 4.0.')

        token_generator = token_generator_registry.get_token_generator(
            token_generator_id)

        if token_generator is None:
            logger.error('Could not find the %s token generator in the '
                         'registry.', token_generator_id)

            raise WebAPITokenGenerationError(
                _('Could not find the token generator to generate the token.'))

        if isinstance(policy, dict):
            policy = json.dumps(policy)

        for attempt in range(max_attempts):
            if 'attempt' in token_info:
                token_info['attempt'] = attempt

            token_string = token_generator.create_token(token_info=token_info)

            try:
                token = self.create(user=user,
                                    token=token_string,
                                    token_generator_id=token_generator_id,
                                    note=note or '',
                                    policy=policy,
                                    expires=expires,
                                    **kwargs)
            except IntegrityError:
                # We hit a collision with the token value. Try again.
                pass
            else:
                webapi_token_created.send(
                    instance=token,
                    auto_generated=auto_generated,
                    sender=type(token),
                )
                return token

        # We hit our limit. The database is too full of tokens.
        logger.error('Unable to generate unique API token for %s after '
                     '%d attempts.',
                     user.username,
                     max_attempts)

        raise WebAPITokenGenerationError(
            _('Could not create a unique API token. Please try again.'))

    def invalidate_tokens(self,
                          users=None,
                          extra_query=None,
                          invalid_reason=''):
        """Invalidate a set of tokens.

        Version Added:
            3.0

        Args:
            users (list of int, optional):
                The list of user IDs whose tokens will be invalidated. By
                default this is set to ``None``, in which case the tokens for
                all users will be invalidated. The set of tokens to invalidate
                can be further filtered down using :py:param`extra_query`.

            extra_query (django.db.models.Q, optional):
                Additional query parameters to filter the set of tokens that
                will be invalidated.

            invalid_reason (str, optional):
                A message indicating why the tokens are invalid. This defaults
                to an empty string.
        """
        q = Q()

        if users:
            q &= Q(user__in=users)

        if extra_query:
            q &= extra_query

        self.filter(q).update(valid=False,
                              invalid_reason=invalid_reason,
                              invalid_date=timezone.now())

    def invalidate_token(self, token, invalid_reason=''):
        """Invalidate the given token.

        Version Added:
            3.0

        Args:
            token (str):
                The token value of the token to invalidate.

            invalid_reason (str, optional):
                A message indicating why the token is invalid. This defaults
                to an empty string.
        """
        self.invalidate_tokens(extra_query=Q(token=token),
                               invalid_reason=invalid_reason)



"""Basic version and package information."""

# Python 3.10+ removed a bunch of aliases from the collections module.
# Unfortunately, Django 1.11 and nose still rely on these. Do this before
# anything else.
import djblets.util.compat.python.collections


# The version of Djblets
#
# This is in the format of:
#
#   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#
VERSION = (3, 1, 0, 'final', 0, True)


def get_version_string():
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%s" % VERSION[2]

    if VERSION[3] != 'final':
        if VERSION[3] == 'rc':
            version += ' RC%s' % VERSION[4]
        else:
            version += ' %s %s' % (VERSION[3], VERSION[4])

    if not is_release():
        version += " (dev)"

    return version


def get_package_version():
    major, minor, micro, tag, release_num, released = VERSION

    version = '%d.%d' % (major, minor)

    if micro:
        version = '%s.%d' % (version, micro)

    if tag != 'final':
        if tag == 'alpha':
            tag = 'a'
        elif tag == 'beta':
            tag = 'b'

        version = '%s%s%s' % (version, tag, release_num)

    return version


def is_release():
    return VERSION[5]


__version_info__ = VERSION[:-1]
__version__ = get_package_version()



"""Condition rule support for applications.

This module contains convenience imports for:

* :py:class:`~djblets.conditions.conditions.Condition`
* :py:class:`~djblets.conditions.conditions.ConditionSet`
"""

from djblets.conditions.conditions import Condition, ConditionSet


__all__ = [
    'Condition',
    'ConditionSet',
]



from django.db.models import Q
from django.test.client import RequestFactory

from djblets.webapi.resources.mixins.queries import APIQueryUtilsMixin
from djblets.testing.testcases import TestCase


class APIQueryUtilsMixinTests(TestCase):
    """Unit tests for djblets.webapi.resources.mixins.queries."""

    def setUp(self):
        self.query_utils = APIQueryUtilsMixin()
        self.request_factory = RequestFactory()

    def test_build_queries_for_int_field_with_lt(self):
        """Testing APIQueryUtilsMixin.build_queries_for_int_field with -lt="""
        request = self.request_factory.get('/api/test/?myfield-lt=10')
        q = self.query_utils.build_queries_for_int_field(request, 'myfield')

        self.assertQEqual(q, Q(myfield__lt=10))

    def test_build_queries_for_int_field_with_lte(self):
        """Testing APIQueryUtilsMixin.build_queries_for_int_field with -lte="""
        request = self.request_factory.get('/api/test/?myfield-lte=10')
        q = self.query_utils.build_queries_for_int_field(request, 'myfield')

        self.assertQEqual(q, Q(myfield__lte=10))

    def test_build_queries_for_int_field_with_gt(self):
        """Testing APIQueryUtilsMixin.build_queries_for_int_field with -gt="""
        request = self.request_factory.get('/api/test/?myfield-gt=10')
        q = self.query_utils.build_queries_for_int_field(request, 'myfield')

        self.assertQEqual(q, Q(myfield__gt=10))

    def test_build_queries_for_int_field_with_gte(self):
        """Testing APIQueryUtilsMixin.build_queries_for_int_field with -gte="""
        request = self.request_factory.get('/api/test/?myfield-gte=10')
        q = self.query_utils.build_queries_for_int_field(request, 'myfield')

        self.assertQEqual(q, Q(myfield__gte=10))

    def test_build_queries_for_int_field_with_query_param_name(self):
        """Testing APIQueryUtilsMixin.build_queries_for_int_field
        with query_param_name
        """
        request = self.request_factory.get('/api/test/?foo-gte=10')
        q = self.query_utils.build_queries_for_int_field(request, 'myfield',
                                                         'foo')

        self.assertQEqual(q, Q(myfield__gte=10))

    def assertQEqual(self, q1, q2):
        self.assertEqual(q1.connector, q2.connector)
        self.assertEqual(q1.children, q2.children)



"""Unit tests for djblets.extensions.hooks.DataGridColumnsHook."""

from kgb import SpyAgency

from djblets.datagrid.grids import Column, DataGrid
from djblets.extensions.extension import Extension
from djblets.extensions.hooks import DataGridColumnsHook
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase


class MyTestExtension(Extension):
    pass


class DataGridColumnsHookTest(SpyAgency, ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.DataGridColumnsHook."""

    extension_class = MyTestExtension

    def test_add_column(self):
        """Testing DataGridColumnsHook registers column"""
        self.spy_on(DataGrid.add_column)

        DataGridColumnsHook(extension=self.extension,
                            datagrid_cls=DataGrid,
                            columns=[Column(id='sandbox')])

        self.assertTrue(DataGrid.add_column.called)

    def test_remove_column(self):
        """Testing DataGridColumnsHook unregisters column"""
        self.spy_on(DataGrid.remove_column)

        hook = DataGridColumnsHook(extension=self.extension,
                                   datagrid_cls=DataGrid,
                                   columns=[Column(id='sandbox2')])

        hook.disable_hook()

        self.assertTrue(DataGrid.remove_column.called)



"""Admin site definitions for extension models."""

from django.contrib import admin

from djblets.extensions.models import RegisteredExtension


class RegisteredExtensionAdmin(admin.ModelAdmin):
    list_display = ('class_name', 'name', 'enabled')


admin.site.register(RegisteredExtension, RegisteredExtensionAdmin)



"""Common consent requirements for use in applications.

This provides some basic definitions with standard requirement IDs that can be
further customized by applications needing to offer consent for services.
"""

from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, gettext

from djblets.privacy.consent.base import BaseConsentRequirement


class BaseGravatarConsentRequirement(BaseConsentRequirement):
    """Base consent requirement for Gravatar usage.

    This supplies a requirement ID, name, default summary, and default data
    use description for Gravatar consent requirements. Subclasses should
    provide their own intent and data use descriptions.
    """

    requirement_id = 'gravatar'
    name = _('Gravatar')
    summary = _("We'd like to use Gravatar.com for your avatars")

    data_use_description = _(
        "Gravatar will receive a one-way hashed version of your e-mail "
        "address. This is not personally identifiable, but could potentially "
        "be used to track you across multiple sites that use Gravatar. "
        "Gravatar only has an avatar for you if you've set one with their "
        "service."
    )


class BaseIntercomConsentRequirement(BaseConsentRequirement):
    """Base consent requirement for Intercom usage.

    This supplies a requirement ID, name, and default summary for Intercom
    consent requirements. Subclasses should provide their own intent and
    data use descriptions.
    """

    requirement_id = 'intercom'
    name = 'Intercom'
    summary = _("We'd like to support and track activity using Intercom")


class PolicyConsentRequirement(BaseConsentRequirement):
    """A consent requirement for asking users to acknowledge policies."""

    requirement_id = 'policies'
    allow_text = _('Accept')
    allowed_text = _('Accepted')
    block_text = _('Reject')
    blocked_text = _('Rejected')

    def __init__(self, privacy_policy_url, terms_of_service_url,
                 site_admin_email=None, reject_instructions=None):
        """Initialize the consent requirement.

        Args:
            privacy_policy_url (unicode):
                The URL to the privacy policy, if applicable.

            terms_of_service (unicode):
                The URL to the terms of service, if applicable.

            site_admin_email (unicode, optional):
                The e-mail address of the site admin. This is only used if
                ``reject_instructions`` is provided.

            reject_instructions (unicode, optional):
                Instructions for how the user should proceed if they do not
                accept the policies.

        """
        if site_admin_email is None and reject_instructions is None:
            raise ValueError('Either reject_instructions or '
                             'site_admin_email must be provided.')

        self.privacy_policy_url = privacy_policy_url
        self.terms_of_service_url = terms_of_service_url
        self.site_admin_email = site_admin_email
        self.reject_instructions = reject_instructions

    @property
    def name(self):
        """The name of the requirement."""
        if self.privacy_policy_url and self.terms_of_service_url:
            return gettext('Privacy Policy and Terms of Service')
        elif self.privacy_policy_url:
            return gettext('Privacy Policy')
        elif self.terms_of_service_url:
            return gettext('Terms of Service')

    @property
    def summary(self):
        """A brief summary of the requirement."""
        if self.privacy_policy_url and self.terms_of_service_url:
            return gettext('Accept the Privacy Policy and Terms of Service')
        elif self.privacy_policy_url:
            return gettext('Accept the Privacy Policy')
        elif self.terms_of_service_url:
            return gettext('Accept the Terms of Service')

    @property
    def intent_description(self):
        """A short description of why the requirement is needed."""
        if self.reject_instructions:
            reject_instructions = self.reject_instructions
        else:
            reject_instructions = (
                gettext('You will not be able to use the service '
                        'without accepting this. If you disagree, '
                        'please <a href="mailto:%s">contact</a> '
                        'your server administrator.')
                % self.site_admin_email
            )

        if self.privacy_policy_url and self.terms_of_service_url:
            return mark_safe(
                gettext('You must accept the <a href="%s">Privacy '
                        'Policy</a> and <a href="%s">Terms of '
                        'Service</a>. %s')
                % (self.privacy_policy_url, self.terms_of_service_url,
                   reject_instructions)
            )
        elif self.privacy_policy_url:
            return mark_safe(
                gettext('You must accept the <a href="%s">Privacy '
                        'Policy</a>. %s')
                % (self.privacy_policy_url, reject_instructions)
            )
        elif self.terms_of_service_url:
            return mark_safe(
                gettext('You must accept the <a href="%s">Terms of '
                        'Service</a>. %s')
                % (self.terms_of_service_url, reject_instructions)
            )



"""Unit tests for djblets.util.templatetags.djblets_email."""

from django.template import Context, Template, TemplateSyntaxError

from djblets.testing.testcases import TagTest, TestCase
from djblets.util.templatetags.djblets_email import quote_text


class CondenseTagTests(TagTest):
    """Unit tests for the {% condense %} template tag."""

    tag_content = 'foo\nbar\n\n\n\n\n\n\nfoobar!'

    def test_with_defaults(self):
        """Testing {% condense %}"""
        t = Template('{% load djblets_email %}'
                     '{% condense %}' +
                     self.tag_content +
                     '{% endcondense %}')

        self.assertHTMLEqual(t.render(Context({})),
                             'foo\nbar\n\n\nfoobar!')

    def test_with_max_indents(self):
        """Testing {% condense %} with custom max_indents"""
        t = Template('{% load djblets_email %}'
                     '{% condense 1 %}' +
                     self.tag_content +
                     '{% endcondense %}')
        self.assertHTMLEqual(t.render(Context({})),
                             'foo\nbar\nfoobar!')


class QuotedEmailTagTests(TagTest):
    """Unit tests for the {% quoted_email %} template tag."""

    def test_basic_usage(self):
        """Testing {% quoted_email %}"""
        t = Template('{% load djblets_email %}'
                     '{% quoted_email template_name %}')

        self.assertEqual(
            t.render(Context({
                'template_name': 'testing/foo.html',
                'foo': 'baz',
                'bar': 'qux',
            })),
            '> baz qux\n'
            '>')

    def test_with_invalid(self):
        """Testing {% quoted_email %} with invalid usage"""
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load djblets_email %}'
                     '{% quoted_email %}content{% end_quoted_email %}')


class QuoteTextFilterTests(TestCase):
    """Unit tests for the {{...|quote_text}} template filter."""

    def test_with_default_level(self):
        """Testing {{...|quote_text}} with default quote level"""
        self.assertEqual(quote_text('foo\nbar'),
                         '> foo\n> bar')

    def testLevel2(self):
        """Testing {{...|quote_text}} with custom quote level"""
        self.assertEqual(quote_text('foo\nbar', 2),
                         '> > foo\n> > bar')



from copy import deepcopy

from django.utils.translation import gettext as _


class JSONPatchError(Exception):
    """An error occurred while patching an object.

    Attributes:
        doc (dict or list):
            The original JSON document the patch was being applied to. This
            won't contain the patched modifications.

        patch (object):
            The patch being applied. The value depends on the type of patch.

        patch_entry_index (int, optional):
            The index of the patch being applied. This can be ``None``
            if the error doesn't apply to a specific entry.
    """

    def __init__(self, msg, doc, patch, patch_entry_index=None):
        """Initialize the error.

        Args:
            msg (unicode):
                The error message.

            doc (dict or list):
                The original JSON document the patch was being applied to. This
                won't contain the patched modifications.
                patch.

            patch (object):
                The patch being applied. The value depends on the patch.

            patch_entry_index (int, optional):
                The index of the patch being applied. This can be ``None``
                if the error doesn't apply to a specific entry.
        """
        super(JSONPatchError, self).__init__(msg)

        self.doc = doc
        self.patch = patch
        self.patch_entry_index = patch_entry_index


class JSONPatchPathError(JSONPatchError):
    """Error with a path in a JSON Patch."""

    def __init__(self, msg, path, **kwargs):
        """Initialize the error.

        Args:
            msg (unicode):
                The error message describing the path failure.

            path (unicode):
                The path that had an error.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent
                class.
        """
        super(JSONPatchPathError, self).__init__(msg, **kwargs)

        self.path = path


class JSONPatchAccessError(JSONPatchPathError):
    """Access error reading from or writing to part of an object."""


class JSONPatchReadAccessError(JSONPatchAccessError):
    """Access error reading from part of an object."""


class JSONPatchWriteAccessError(JSONPatchAccessError):
    """Access error writing to part of an object."""


class JSONPatchTestError(JSONPatchPathError):
    """Test condition failed when applying a JSON Patch."""


class JSONPointerError(Exception):
    """Base class for JSON Pointer errors."""


class JSONPointerSyntaxError(JSONPointerError):
    """Syntax error in a JSON Pointer path."""


class JSONPointerLookupError(JSONPointerError):
    """Error looking up data from a JSON Pointer path."""

    def __init__(self, msg, parent, token, token_index, tokens):
        """Initialize the error.

        Args:
            msg (unicode):
                The error message.

            parent (object):
                The parent object. This may be a dictionary or a list.

            token (unicode):
                The last resolvable token in the path. This will be ``None``
                if the first token failed.

            token_index (int):
                The 0-based index of the last resolvable token in the path.
                This will be ``None`` if the first token failed.

            tokens (list of unicode):
                The list of tokens comprising the full path.
        """
        super(JSONPointerLookupError, self).__init__(msg)

        self.parent = parent
        self.token = token
        self.token_index = token_index
        self.tokens = tokens


class JSONPointerEndOfList(object):
    """A representation of the end of a list.

    This is used by the JSON Pointer functions to represent the very end of
    a list (not the last item). This is used primarily when specifying an
    insertion point, with this value repersenting appending to a list.
    """

    def __init__(self, json_list):
        """Initialize the object.

        Args:
            json_list (list):
                The list this object represents.
        """
        self.json_list = json_list

    def __eq__(self, other_list):
        """Return whether two instances are equal.

        Instances are equal if both of their lists are equal.

        Args:
            other_list (JSONPointerEndOfList):
                The other instance to compare to for equality.

        Returns:
            bool:
            ``True`` if the two lists are equal.
        """
        try:
            return self.json_list == other_list.json_list
        except AttributeError:
            return False

    def __repr__(self):
        """Return a string representation of the instance.

        Returns:
            unicode:
            A string representation of the instance.
        """
        return 'JSONPointerEndOfList<%r>' % self.json_list


def json_merge_patch(doc, patch, can_write_key_func=None):
    """Apply a JSON Merge Patch to a value.

    This is an implementation of the proposed JSON Merge Patch standard
    (:rfc:`7396`), which is a simple algorithm for applying changes to a
    JSON-compatible data structure.

    This will attempt to merge the contents of the ``patch`` object into the
    ``doc`` object as best as possible, using the following rules:

    * If a key in ``patch`` matches a key in ``doc``, and the value is
      ``None``, the key in ``doc`` is removed.

    * If a key in ``patch`` matches a key in ``doc``, and the value is a
      dictionary, this method will be called on the values in each and the
      result will be stored as the new value.

    * If a key in ``patch`` matches a key in ``doc``, and the value is not a
      dictionary, the value from ``patch`` will replace the value in
      ``doc``.

    * If ``patch`` is a dictionary but ``doc`` is not, then ``patch`` will
      be returned directly.

    * If ``patch`` is a dictionary but ``doc`` is not, then ``doc`` will
      be discarded and the JSON Merge Patch will be applied to a new empty
      dictionary.

    * If ``patch`` is not a dictionary, then ``patch`` will be returned
      directly.

    Args:
        doc (object):
            The JSON-compatible document object the patch is being applied to.
            This will usually be a :js:class:`dict`.

        patch (object):
            The JSON-compatible patch to apply.

        can_write_key_func (callable, optional):
            An optional function to call for each key to determine if the
            key can be written in the new dictionary. This must take the
            following form:

            .. code-block:: python

               def can_write_key(doc, patch, path, **kwargs):
                   ...

            This must return a boolean indicating if the key can be written, or
            may raise a :py:class:`JSONPatchError` to abort the patching
            process. If the function returns ``False``, the key will simply be
            skipped.

    Returns:
        object:
        The resulting object. This will be the same type as ``patch``.

    Raises:
        JSONPatchError:
            There was an error patching the document. This is only raised by
            ``can_write_key_func`` implementations.

        ValueError:
            A provided parameter was incorrect.
    """
    def _merge_patch(cur_doc, cur_patch, parent_path=()):
        if isinstance(cur_patch, dict):
            if isinstance(cur_doc, dict):
                new_doc = cur_doc.copy()
            else:
                # Override the contents of the doc with a new, empty
                # dictionary, which will then be populated.
                new_doc = {}

            for key, value in cur_patch.items():
                path = parent_path + (key,)

                if (can_write_key_func and
                    not can_write_key_func(doc=doc,
                                           patch=patch,
                                           path=path)):
                    # We can't write this key. We'll skip it, since the
                    # function didn't raise an exception.
                    continue

                if value is None:
                    new_doc.pop(key, None)
                else:
                    new_doc[key] = _merge_patch(new_doc.get(key), value, path)

            return new_doc
        else:
            return cur_patch

    if can_write_key_func and not callable(can_write_key_func):
        raise ValueError('can_write_key_func must be callable')

    return _merge_patch(doc, patch)


def _json_patch_add(obj, key, value, doc, patch, patch_entry_index, path):
    """Add a value to a JSON object from a patch.

    This will add a value to a dictionary or list based on the rules of a
    JSON Patch operation.

    If working on a list, ``-`` is considered a valid key, which will result
    in the object being appended to the list.

    Args:
        obj (dict or list):
            The object that the value is being added to.

        key (unicode):
            The key (for dictionaries) or index (for lists) for the value.
            This also allows for a special ``-`` for lists, indicating that
            the value should be appended to the list.

        value (object):
            The value to add to the object.

        doc (dict or list):
            The root JSON document being patched.

        patch (dict or list):
            The main patch being applied.

        patch_entry_index (int):
            The patch entry index for this operation.

        path (unicode):
            The path being updated.

    Raises:
        JSONPatchError:
            There was an error adding the value to the object. Details are
            in the message.
    """
    if isinstance(obj, list):
        if key == '-':
            obj.append(value)
        else:
            # This is guaranteed to be an integer here. The JSON Pointer code
            # checks for this.
            key = int(key)

            if key > len(obj):
                raise JSONPatchPathError(
                    _('Cannot insert into index %(index)d in path "%(path)s" '
                      'for patch entry %(entry)d')
                    % {
                        'index': key,
                        'path': path,
                        'entry': patch_entry_index,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=patch_entry_index,
                    path=path)

            obj.insert(key, value)
    elif isinstance(obj, dict):
        obj[key] = value
    else:
        raise JSONPatchPathError(
            _('Unable to add key "%(key)s" to a non-dictionary/list in path '
              '"%(path)s" for patch entry %(entry)d')
            % {
                'key': key,
                'path': path,
                'entry': patch_entry_index,
            },
            doc=doc,
            patch=patch,
            patch_entry_index=patch_entry_index,
            path=path)


def _json_patch_remove(obj, key, doc, patch, patch_entry_index, path):
    """Remove a value from a JSON object from a patch.

    This will remove the value from a dictionary or list based on the rules of
    a JSON Patch operation.

    Args:
        obj (dict or list):
            The object that the value is being added to.

        key (unicode):
            The key (either a dictionary key or a list index) to remove.

        doc (dict or list):
            The root JSON document being patched.

        patch (dict or list):
            The main patch being applied.

        patch_entry_index (int):
            The patch entry index for this operation.

        path (unicode):
            The path being updated.

    Raises:
        JSONPatchError:
            There was an error adding the value to the object. Details are
            in the message.
    """
    if isinstance(obj, list):
        # This is guaranteed to be an integer here. The JSON Pointer code
        # checks for this.
        key = int(key)

    # Even though this works with lists, IndexErrors should never occur, as
    # bounds checks will be handled by JSON Pointer.
    try:
        del obj[key]
    except KeyError:
        raise JSONPatchPathError(
            _('Cannot remove non-existent key "%(key)s" in path "%(path)s" '
              'for patch entry %(entry)d')
            % {
                'key': key,
                'path': path,
                'entry': patch_entry_index,
            },
            doc=doc,
            patch=patch,
            patch_entry_index=patch_entry_index,
            path=path)


def json_patch(doc, patch, can_read_key_func=None, can_write_key_func=None):
    """Apply a JSON Patch to a value.

    A JSON Patch (:rfc:`6902`), similar to a JSON Merge Patch, is used to make
    changes to an existing JSON-compatible data structure. JSON Patches are
    composed of a list of operations which can add a new value, remove a value,
    replace an existing value, move a value, copy a value, or test that a path
    has a given value. All operations must pass for a patch to apply.

    A full description of the operations are available in the RFC or from
    http://jsonpatch.com/.

    Args:
        doc (dict or list):
            The root JSON document to apply the patch to. This will not be
            modified itself.

        patch (list of dict):
            The list of operations to apply to the JSON document.

        can_read_key_func (callable, optional):
            An optional function to call for each path to determine if the
            path can be read from the document. This is in the following form:

            .. code-block:: python

               def can_read_key(doc, patch, patch_entry, path, **kwargs):
                   ...

            It must return a boolean indicating if the key can be written.

        can_write_key_func (callable, optional):
            An optional function to call for each path to determine if the
            path can be written to the document. This takes the same form as
            ``can_read_key_func``.

            If not provided, it will default to ``can_read_key_func``.

    Returns:
        dict or list:
        The resulting JSON document after the patch is applied.

    Raises:
        JSONPatchError:
            An error occurred patching the JSON document.
    """
    if not isinstance(patch, list):
        raise JSONPatchError(
            'The patch must be a list of operations to perform',
            doc=doc,
            patch=patch)

    new_doc = deepcopy(doc)

    if can_read_key_func and not callable(can_read_key_func):
        raise ValueError('can_read_key_func must be callable')

    if not can_write_key_func:
        can_write_key_func = can_read_key_func
    elif not callable(can_write_key_func):
        raise ValueError('can_write_key_func must be callable')

    for i, patch_entry in enumerate(patch):
        if not isinstance(patch_entry, dict):
            raise JSONPatchError(
                _('Patch entry %(entry)d must be a dictionary instead of '
                  '%(type)s')
                % {
                    'type': type(patch_entry).__name__,
                    'entry': i,
                },
                doc=doc,
                patch=patch,
                patch_entry_index=i)

        value = None
        from_path = None
        from_path_info = None
        from_all_tokens = None

        # Pull some common, simple variables out of the patch entry. Some are
        # only common to specific operations.
        try:
            op = patch_entry['op']
            path = patch_entry['path']

            if op in ('add', 'replace', 'test'):
                value = patch_entry['value']
            elif op in ('copy', 'move'):
                from_path = patch_entry['from']
        except KeyError as e:
            raise JSONPatchError(
                _('Missing key "%(key)s" for patch entry %(entry)d')
                % {
                    'key': e.args[0],
                    'entry': i,
                },
                doc=doc,
                patch=patch,
                patch_entry_index=i)

        # All operations have a path. Parse it and check for syntax errors.
        # We'll check if the path was fully resolved (or if we care) next.
        try:
            path_info = json_get_pointer_info(new_doc, path)
        except JSONPointerSyntaxError as e:
            raise JSONPatchPathError(
                _('Syntax error in path "%(path)s" for patch entry %(entry)d: '
                  '%(error)s')
                % {
                    'path': path,
                    'entry': i,
                    'error': e,
                },
                doc=doc,
                patch=patch,
                patch_entry_index=i,
                path=path)

        all_tokens = tuple(path_info['all_tokens'])

        # Check for path validity. The "remove" and "test" operations
        # require a fully-resolvable path. Others allow the last part of
        # the path to be unresolved, as those can be target locations for a
        # write operation.
        lookup_error = path_info['lookup_error']

        if lookup_error:
            if (op in ('remove', 'test') or
                len(path_info['unresolved_tokens']) > 1):
                raise JSONPatchPathError(
                    _('Invalid path "%(path)s" for patch entry %(entry)d: '
                      '%(error)s')
                    % {
                        'path': path,
                        'entry': i,
                        'error': lookup_error,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=path)
        else:
            # If the caller referenced '-' for the end of an array, make sure
            # it's an operation where that's allowed.
            if (isinstance(path_info['value'], JSONPointerEndOfList) and
                op not in ('add', 'copy', 'move')):
                raise JSONPatchPathError(
                    _('Cannot perform operation "%(op)s" on end of list at '
                      '"%(path)s" for patch entry %(entry)d')
                    % {
                        'op': op,
                        'path': path,
                        'entry': i,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=path)

        # If the caller supplied can_read_key_func/can_write_key_func, check
        # if the key can be read from or written to. If not, return an error.
        if op == 'test':
            if (can_read_key_func and
                not can_read_key_func(doc=new_doc,
                                      patch_entry=patch_entry,
                                      path=all_tokens)):
                raise JSONPatchReadAccessError(
                    _('Cannot read from path "%(path)s" for patch entry '
                      '%(entry)d')
                    % {
                        'path': path,
                        'entry': i,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=path)
        else:
            if (can_write_key_func and
                not can_write_key_func(doc=new_doc,
                                       patch_entry=patch_entry,
                                       path=all_tokens)):
                raise JSONPatchWriteAccessError(
                    _('Cannot write to path "%(path)s" for patch entry '
                      '%(entry)d')
                    % {
                        'path': path,
                        'entry': i,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=path)

        # copy/move operations have a second path to consider: the "from" path.
        # These require a new set of validity and access checks.
        if op in ('copy', 'move'):
            if from_path == path:
                # If the two paths are the same, there's nothing to do. It's
                # not even worth checking for access, since this is a no-op.
                continue

            # "copy" and "move" operations require a full "from" path.
            # First parse and check for syntax errors.
            try:
                from_path_info = json_get_pointer_info(new_doc,
                                                       from_path)
            except JSONPointerSyntaxError as e:
                raise JSONPatchPathError(
                    _('Syntax error in from path "%(path)s" for patch '
                      'entry %(entry)d: %(error)s')
                    % {
                        'path': from_path,
                        'entry': i,
                        'error': e,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=from_path)

            # Now check to see if the path can be resolved. Partially-resolved
            # "from" paths aren't allowed, unlike standard paths for some
            # other operations.
            lookup_error = from_path_info['lookup_error']

            if lookup_error:
                raise JSONPatchPathError(
                    _('Invalid from path "%(path)s" for patch entry '
                      '%(entry)d: %(error)s')
                    % {
                        'path': from_path,
                        'entry': i,
                        'error': lookup_error,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=from_path)

            # The end of list ('-') is not a valid "from" location for these
            # operations, so error out if we see it.
            if isinstance(from_path_info['value'], JSONPointerEndOfList):
                raise JSONPatchPathError(
                    _('Cannot perform operation "%(op)s" from end of list at '
                      '"%(path)s" for patch entry %(entry)d')
                    % {
                        'op': op,
                        'path': from_path,
                        'entry': i,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=from_path)

            from_all_tokens = tuple(from_path_info['all_tokens'])

            if op == 'move':
                # For move operations, we need to check if we can write to
                # the old location, since we'll be removing it.
                if (can_write_key_func and
                    not can_write_key_func(doc=new_doc,
                                           patch_entry=patch_entry,
                                           path=from_all_tokens)):
                    raise JSONPatchWriteAccessError(
                        _('Cannot write to path "%(path)s" for patch entry '
                          '%(entry)d')
                        % {
                            'path': from_path,
                            'entry': i,
                        },
                        doc=doc,
                        patch=patch,
                        patch_entry_index=i,
                        path=from_path)
            else:
                # For copy operations, we just need to check if we can read
                # from the old location.
                if (can_read_key_func and
                    not can_read_key_func(doc=new_doc,
                                          patch_entry=patch_entry,
                                          path=from_all_tokens)):
                    raise JSONPatchReadAccessError(
                        _('Cannot read from path "%(path)s" for patch entry '
                          '%(entry)d')
                        % {
                            'path': from_path,
                            'entry': i,
                        },
                        doc=doc,
                        patch=patch,
                        patch_entry_index=i,
                        path=from_path)

        parent = path_info['parent']

        try:
            last_token = all_tokens[-1]
        except IndexError:
            last_token = None

        # We can now begin processing the operation for this entry.
        if op == 'add':
            _json_patch_add(obj=parent,
                            key=last_token,
                            value=value,
                            doc=doc,
                            patch=patch,
                            patch_entry_index=i,
                            path=path)
        elif op == 'remove':
            _json_patch_remove(obj=parent,
                               key=last_token,
                               doc=doc,
                               patch=patch,
                               patch_entry_index=i,
                               path=path)
        elif op == 'replace':
            if parent is None:
                # If there's no parent, the entire document will be replaced.
                new_doc = value
            else:
                _json_patch_remove(obj=parent,
                                   key=last_token,
                                   doc=doc,
                                   patch=patch,
                                   patch_entry_index=i,
                                   path=path)
                _json_patch_add(obj=parent,
                                key=last_token,
                                value=value,
                                doc=doc,
                                patch=patch,
                                patch_entry_index=i,
                                path=path)
        elif op == 'copy':
            _json_patch_add(obj=parent,
                            key=last_token,
                            value=deepcopy(from_path_info['value']),
                            doc=doc,
                            patch=patch,
                            patch_entry_index=i,
                            path=path)
        elif op == 'move':
            num_from_tokens = len(from_all_tokens)

            # Make sure this isn't attempting to move a path into its own
            # tree somewhere.
            if (len(all_tokens) > num_from_tokens and
                all_tokens[:num_from_tokens] == from_all_tokens):
                raise JSONPatchPathError(
                    _('Cannot move values into their own children at patch '
                      'entry %(entry_id)d')
                    % {
                        'entry_id': i,
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=from_path)

            # Per spec, moves must be implemented as a remove followed
            # immediately by an add.
            _json_patch_remove(obj=from_path_info['parent'],
                               key=from_all_tokens[-1],
                               doc=doc,
                               patch=patch,
                               patch_entry_index=i,
                               path=from_path)
            _json_patch_add(obj=parent,
                            key=last_token,
                            value=from_path_info['value'],
                            doc=doc,
                            patch=patch,
                            patch_entry_index=i,
                            path=path)
        elif op == 'test':
            # We already know the path is valid, so we just need to check the
            # value. If it doesn't match, abort the patching process.
            if path_info['value'] != value:
                raise JSONPatchTestError(
                    _('Test failed for path "%(path)s" at patch entry '
                      '%(entry_id)d. Expected %(expected_value)r and got '
                      '%(found_value)r.')
                    % {
                        'path': path,
                        'entry_id': i,
                        'expected_value': value,
                        'found_value': path_info['value'],
                    },
                    doc=doc,
                    patch=patch,
                    patch_entry_index=i,
                    path=path)

    return new_doc


def json_get_pointer_info(obj, path):
    """Return information from a JSON object based on a JSON Pointer path.

    JSON Pointers are a standard way of specifying a path to data within a
    JSON object. This method takes a JSON Pointer path and returns information
    from the given object based on that path.

    Pointer paths consist of dictionary keys, array indexes, or a special ``-``
    token (representing the end of a list, after the last item) separated by
    ``/`` tokens. There are also two special escaped values: ``~0``,
    representing a ``~`` character, and ``~1``, representing a ``/`` character.

    Paths must start with ``/``, or must be an empty string (which will match
    the provided object). If the path has a trailing ``/``, then the final
    token is actually an empty string (matching an empty key in a dictionary).

    If the Pointer does not resolve to a complete path (for instance, a key
    specified in the path is missing), the resulting information will return
    the keys that could be resolved, keys that could not be resolved, the
    object where it left off, and an error message. This allows implementations
    to determine which part of a path does not yet exist, potentially for the
    purpose of inserting data at that key in the path.

    Args:
        obj (object or list):
            The object or list representing the starting object for the path.

        path (unicode):
            The Pointer path for the lookup.

    Returns:
        dict:
        Information about the object and what the path was able to match. This
        has the following keys:

        ``value`` (:py:class:`object`):
            The resulting value from the path, if the path was fully resolved.

            This will be a :py:class:`JSONPointerEndOfList` if the last part of
            the path was a ``-`` token (representing the very end of
            a list).

        ``tokens`` (:py:class:`list`):
            The normalized (unescaped) list of path tokens that comprise the
            full path.

        ``parent`` (:py:class:`object`):
            The parent object for either the most-recently resolved value.

        ``resolved`` (:py:class:`list`):
            The list of resolved objects from the original JSON object. This
            will contain each key, array item, or other value that was found,
            in path order.

        ``lookup_error`` (:py:class:`unicode`):
            The error message, if any, if failing to resolve the full path.

    Raises:
        JSONPointerSyntaxError:
            There was a syntax error with a token in the path.
    """
    if path != '' and not path.startswith('/'):
        raise JSONPointerSyntaxError(
            _('Paths must either be empty or start with a "/"'))

    # Split the path into segments, trimming off the first entry (the root
    # object). If path is empty, the split will be a no-op, and we'll end up
    # operating on the root object itself.
    tokens = path.split('/')[1:]

    # Decode the special values for "/" and "~". The decode order here is
    # important.
    norm_tokens = [
        token.replace('~1', '/').replace('~0', '~')
        for token in tokens
    ]

    resolved = [obj]
    lookup_error = None
    parent = None

    for i, token in enumerate(norm_tokens):
        parent = obj

        if isinstance(obj, dict):
            try:
                obj = obj[token]
            except KeyError:
                lookup_error = (
                    _('Dictionary key "%(key)s" not found in "%(path)s"')
                    % {
                        'key': token,
                        'path': '/%s' % '/'.join(tokens[:i]),
                    }
                )
        elif isinstance(obj, list):
            if token == '-':
                obj = JSONPointerEndOfList(obj)
            else:
                if token != '0' and token.startswith('0'):
                    raise JSONPointerSyntaxError(
                        _('List index "%s" must not begin with "0"') % token)

                try:
                    token = int(token)
                except ValueError:
                    raise JSONPointerSyntaxError(
                        _('%(index)r is not a valid list index in "%(path)s"')
                        % {
                            'index': token,
                            'path': '/%s' % '/'.join(tokens[:i]),
                        })

                if token < 0:
                    raise JSONPointerSyntaxError(
                        _('Negative indexes into lists are not allowed'))

                try:
                    obj = obj[token]
                except IndexError:
                    lookup_error = (
                        _('%(index)d is outside the list in "%(path)s"')
                        % {
                            'index': token,
                            'path': '/%s' % '/'.join(tokens[:i]),
                        }
                    )
        else:
            lookup_error = (
                _('Cannot resolve path within unsupported type "%(type)s" at '
                  '"%(path)s"')
                % {
                    'type': type(obj).__name__,
                    'path': '/%s' % '/'.join(tokens[:i]),
                }
            )

        if lookup_error:
            obj = None
            break

        resolved.append(obj)

    return {
        'value': obj,
        'parent': parent,
        'all_tokens': norm_tokens,
        'resolved_values': resolved,
        'resolved_tokens': norm_tokens[:len(resolved) - 1],
        'unresolved_tokens': norm_tokens[len(resolved) - 1:],
        'lookup_error': lookup_error,
    }


def json_resolve_pointer(obj, path):
    """Return the value from a JSON object based on a JSON Pointer path.

    See :py:func:`json_get_pointer_info` for information on how a Pointer
    path is constructed. Unlike that function, this requires a fully-resolved
    path.

    Args:
        obj (object or list):
            The object or list representing the starting object for the path.

        path (unicode):
            The Pointer path for the lookup.

    Returns:
        object:
        The resulting value from the object, based on the path.

        This will be a :py:class:`JSONPointerEndOfList` if the last part of
        the path was a ``-`` token (representing the very end of
        a list).

    Raises:
        JSONPointerLookupError:
            The path could not be fully resolved.

        JSONPointerSyntaxError:
            There was a syntax error with a token in the path.
    """
    info = json_get_pointer_info(obj, path)
    lookup_error = info['lookup_error']

    if lookup_error:
        tokens = info['all_tokens']
        parent = info['parent']

        if parent is None:
            token_index = None
            token = None
        else:
            token_index = len(info['resolved_tokens']) - 1
            token = tokens[token_index]

        raise JSONPointerLookupError(lookup_error,
                                     parent=parent,
                                     token=token,
                                     token_index=token_index,
                                     tokens=tokens)

    return info['value']



"""An avatar service for providing Gravatars."""

from django.utils.html import mark_safe

from djblets.avatars.services.base import AvatarService
from djblets.gravatars import get_gravatar_url_for_email
from djblets.privacy.consent.common import BaseGravatarConsentRequirement


class GravatarService(AvatarService):
    """An avatar service for providing Gravatars."""

    avatar_service_id = 'gravatar'
    name = 'Gravatar'

    consent_requirement_id = BaseGravatarConsentRequirement.requirement_id

    def get_avatar_urls_uncached(self, user, size):
        """Return the Gravatar URLs for the requested user.

        Args:
            user (django.contrib.auth.models.User):
                The user whose avatar URLs are to be fetched.

            size (int):
                The size (in pixels) the avatar is to be rendered at.

        Returns
            dict:
            A dictionary containing the URLs of the user's avatars at normal-
            and high-DPI.
        """
        return {
            '%dx' % resolution: mark_safe(get_gravatar_url_for_email(
                email=user.email, size=(size * resolution)))
            for resolution in (1, 2, 3)
        }

    def get_etag_data(self, user):
        """Return the ETag data for the user's avatar.

        Args:
            user (django.contrib.auth.models.User):
                The user.

        Returns:
            list of unicode:
            The uniquely identifying information for the user's avatar.
        """
        return [self.avatar_service_id, user.email]



from django.dispatch import Signal


#: Emitted when a SiteConfiguration has loaded.
#:
#: This can be used by callers that depend on
#: :py:class:`~djblets.siteconfig.models.SiteConfiguration` to handle reloading
#: or recomputing data from settings that may have changed in another process
#: or server.
#:
#: Args:
#:     siteconfig (djblets.siteconfig.models.SiteConfiguration)
#:         The site configuration that has been loaded.
#:
#:     old_siteconfig (djblets.siteconfig.models.SiteConfiguration)
#:         The old site configuration. The caller can compare the settings
#:         between the new one and this one to see if it needs to handle
#:         anything.
siteconfig_reloaded = Signal()



"""Generator for tokens that contain an identifiable prefix and checksum.

Version Added:
    3.0
"""

import re
import secrets
import string
import zlib
from typing import Any, Dict, List

from django.utils.translation import gettext_lazy as _

from djblets.secrets.token_generators.base import BaseTokenGenerator


class VendorChecksumTokenGenerator(BaseTokenGenerator):
    """Generator for tokens that contain an identifiable prefix and checksum.

    Tokens of this format contain a prefix for identification (such as
    the vendor or type) and a checksum.

    Version Added:
        3.0
    """

    token_generator_id = 'vendor_checksum'

    #: The character set to use for encoding the token.
    #:
    #: Type:
    #:     str
    CHARSET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'

    #: The length of the checksum portion of the token.
    #:
    #: Type:
    #:     int
    CHECKSUM_LENGTH = 6

    #: A regex for validating the characters of a token.
    #:
    #: Type:
    #:     str
    HASH_RE = r'^_[0-9A-Za-z]+$'

    #: The total length of the token.
    #:
    #: Type:
    #:     int
    TOKEN_LENGTH = 255

    def create_token(
        self,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> str:
        """Create a token that contains a prefix and checksum.

        Args:
            token_info (dict):
                A dictionary that must contain the following information about
                the token:

                Keys:
                    token_type (str):
                        The token prefix to set.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            str:
            The token value.

        Raises:
            KeyError:
                The ``token_info`` dictionary is missing a required key.
        """
        self._check_token_info(token_info)

        charset = self.CHARSET

        token_type = token_info['token_type']
        token_type_length = len(token_type)
        token_entropy_length = (self.TOKEN_LENGTH - self.CHECKSUM_LENGTH -
                                token_type_length - 1)

        entropy_data = ''.join(
            secrets.choice(charset)
            for _i in range(token_entropy_length)
        )
        checksum = self._generate_checksum(entropy_data)

        return '%s_%s%s' % (token_type, entropy_data, checksum)

    def validate_token(
        self,
        token: str,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> bool:
        """Return whether the token is a valid token from this generator.

        Args:
            token (str):
                The token to validate.

            token_info (dict):
                A dictionary that contains the following information about the
                token:

                Keys:
                    token_type (str):
                        The prefix that the token must have.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bool:
            Whether the token is a valid token from this generator.
        """
        try:
            self._check_token_info(token_info)
        except KeyError:
            return False

        token_type = token_info['token_type']
        type_length = len(token_type)
        checksum_indice = -self.CHECKSUM_LENGTH
        token_checksum = token[-self.CHECKSUM_LENGTH:]
        checksum = self._generate_checksum(
            token[type_length + 1:checksum_indice])

        # Djblets 3.0 generated token checksums using an incorrect
        # base62-encoding, which resulted in capital and lowercase letters
        # being swapped. We check against checksum.swapcase() to catch those.
        return (len(token) == 255 and
                token.startswith(token_type) and
                re.match(self.HASH_RE, token[type_length:]) is not None and
                (token_checksum == checksum or
                 token_checksum == checksum.swapcase()))

    def _base62_encode(
        self,
        num: int,
    ) -> str:
        """Encode the number using Base62.

        Args:
            num (int):
                The number to encode.

        Returns:
            str:
            The encoded number.
        """
        charset = self.CHARSET

        if num == 0:
            return charset[0]

        result: List[str] = []

        while num:
            num, remainder = divmod(num, 62)
            result.append(charset[remainder])

        return ''.join(reversed(result))

    def _check_token_info(
        self,
        token_info: Dict[str, Any],
    ) -> None:
        """Checks if all required keys are in the token info dictionary.

        Args:
            token_info (dict):
                A dictionary that information about a token. The required
                keys are:

                Keys:
                    token_type (str):
                        The token vendor/type prefix.

        Raises:
            KeyError:
                The ``token_info`` dictionary is missing a required key.
        """
        try:
            token_info['token_type']
        except KeyError as e:
            raise KeyError(
                _('The token_info dictionary must contain a %s key.')
                % e.args[0])

    def _generate_checksum(self, data):
        """Generate a base62-encoded CRC32 checksum for the given data.

        Args:
            data (str):
                The data to generate a checksum for.

        Returns:
            str:
            The base62-encoded CRC32 checksum.
        """
        checksum_data = zlib.crc32(data.encode('utf-8')) & 0xFFFFFFFF

        return self._base62_encode(checksum_data).zfill(self.CHECKSUM_LENGTH)



"""Miscellaneous, useful decorators."""

from functools import update_wrapper, wraps
from inspect import getfullargspec

from django import template
from django.template import TemplateSyntaxError, Variable
from django.utils.functional import cached_property as django_cached_property


# The decorator decorator.  This is copyright unknown, verbatim from
# http://wiki.python.org/moin/PythonDecoratorLibrary
def simple_decorator(decorator):
    """This decorator can be used to turn simple functions
       into well-behaved decorators, so long as the decorators
       are fairly simple. If a decorator expects a function and
       returns a function (no descriptors), and if it doesn't
       modify function attributes or docstring, then it is
       eligible to use this. Simply apply @simple_decorator to
       your decorator and it will automatically preserve the
       docstring and function attributes of functions to which
       it is applied."""
    def new_decorator(f):
        g = decorator(f)
        update_wrapper(g, f)
        return g
    # Now a few lines needed to make simple_decorator itself
    # be a well-behaved decorator.
    update_wrapper(new_decorator, decorator)
    return new_decorator


def augment_method_from(klass):
    """Augments a class's method with new decorators or documentation.

    This is useful when a class needs to add new decorators or new
    documentation to a parent class's method, without changing the behavior
    or burying the existing decorators.

    The methods using this decorator can provide code to run at the end of
    the parent function. Usually, though, it will just have an empty body
    of ``pass``.
    """
    def _dec(func):
        def _call(*args, **kwargs):
            try:
                f = augmented_func(*args, **kwargs)
            finally:
                func(*args, **kwargs)

            return f

        augmented_func = getattr(klass, func.__name__)

        _call.__name__ = func.__name__
        _call.__doc__ = func.__doc__ or augmented_func.__doc__
        _call.__dict__.update(augmented_func.__dict__)
        _call.__dict__.update(func.__dict__)

        real_func = _call.__dict__.get('_augmented_func', augmented_func)
        _call.__dict__['_augmented_func'] = real_func

        return _call

    return _dec


def blocktag(*args, **kwargs):
    """Creates a block template tag with beginning/end tags.

    This does all the hard work of creating a template tag that can
    parse the arguments passed in and then parse all nodes between a
    beginning and end tag (such as myblock/endmyblock).

    By default, the end tag is prefixed with "end", but that can be
    changed by passing ``end_prefix="end_"`` or similar to @blocktag.

    blocktag will call the wrapped function with `context`  and `nodelist`
    parameters, as well as any parameters passed to the tag. It will
    also ensure that a proper error is raised if too many or too few
    parameters are passed.

    Args:
        end_prefix (unicode, optional):
            The prefix for the end tag. This defaults to ``'end'``, but
            template tags using underscores in the name might want to change
            this to ``'end_'``.

        resolve_vars (bool, optional):
            Whether to automatically resolve all variables provided to the
            template tag. By default, variables are resolved. Template tags
            can turn this off if they want to handle variable parsing
            manually.

    Returns:
        callable:
        The resulting template tag function.

    Example:
        .. code-block:: python

            @register.tag
            @blocktag
            def divify(context, nodelist, div_id=None):
                s = "<div"

                if div_id:
                    s += " id='%s'" % div_id

                return s + ">" + nodelist.render(context) + "</div>"
    """
    class BlockTagNode(template.Node):
        def __init__(self, tag_name, tag_func, nodelist, args):
            self.tag_name = tag_name
            self.tag_func = tag_func
            self.nodelist = nodelist
            self.args = args

        def render(self, context):
            if kwargs.get('resolve_vars', True):
                args = [Variable(var).resolve(context) for var in self.args]
            else:
                args = self.args

            return self.tag_func(context, self.nodelist, *args)

    def _blocktag_func(tag_func):
        def _setup_tag(parser, token):
            bits = token.split_contents()
            tag_name = bits[0]
            del(bits[0])

            argspec = getfullargspec(tag_func)
            params = argspec.args
            varargs = argspec.varargs
            defaults = argspec.defaults

            max_args = len(params) - 2  # Ignore context and nodelist
            min_args = max_args - len(defaults or [])

            if len(bits) < min_args or (not varargs and len(bits) > max_args):
                if not varargs and min_args == max_args:
                    raise TemplateSyntaxError(
                        "%r tag takes %d arguments." % (tag_name, min_args))
                else:
                    raise TemplateSyntaxError(
                        "%r tag takes %d to %d arguments, got %d." %
                        (tag_name, min_args, max_args, len(bits)))

            nodelist = parser.parse((('%s%s' % (end_prefix, tag_name)),))
            parser.delete_first_token()
            return BlockTagNode(tag_name, tag_func, nodelist, bits)

        update_wrapper(_setup_tag, tag_func)

        return _setup_tag

    end_prefix = kwargs.get('end_prefix', 'end')

    if len(args) == 1 and callable(args[0]):
        # This is being called in the @blocktag form.
        return _blocktag_func(args[0])
    else:
        # This is being called in the @blocktag(...) form.
        return _blocktag_func


class cached_property(django_cached_property):
    """Decorator for creating a read-only property that caches a value.

    This is a drop-in replacement for Django's
    :py:class:`~django.utils.functional.cached_property` that retains the
    docstring and attributes of the original method.

    While Django 1.8+ does retain the docstring, it does not retain the
    attributes.
    """

    def __init__(self, func):
        """Initialize the property.

        Args:
            func (callable):
                The function that will be called when this property is
                accessed. The property will have its name, documentation,
                and other attributes.
        """
        super(cached_property, self).__init__(func)

        update_wrapper(self, func)


def optional_decorator(decorator, predicate):
    """Optionally apply a decorator given a predicate function.

    Args:
        decorator (callable):
            The decorator to conditionally apply.

        predicate (callable):
            The predicate used to determine when ``decorator`` should be used.
            The arguments provided to the decorated function will be passed to
            this function.

    Returns:
        callable:
        A decorator that, when applied to a function, will call the function
        decorated with ``decorator`` when ``predicate()`` returns ``True``.
        Otherwise it will call the original function.

    Example:
        .. code-block:: python

           from djblets.util.decorators import (optional_decorator,
                                                simple_decorator)

           def predicate(verbose)
               return verbose

           @simple_decorator
           def decorator(f):
               def decorated(verbose):
                   print('Hello from the decorator')
                   return f(verbose)

           @optional_decorator(decorator, predicate)
           def f(x):
                print('Hello from f()')

           # Prints "Hello from the decorator" and "Hello from f()"
           f(verbose=True)

           # Prints "Hello from f()"
           f(verbose=False)
    """
    def _decorator(view):
        with_decorator = decorator(view)

        @wraps(view)
        def decorated(*args, **kwargs):
            if predicate(*args, **kwargs):
                return with_decorator(*args, **kwargs)

            return view(*args, **kwargs)

        return decorated

    return _decorator



"""Import utilities for registries."""

from importlib import import_module

from django.utils.functional import SimpleLazyObject


def lazy_import_registry(module_path, registry_name, **kwargs):
    """Lazily import and construct a registry on access.

    This is useful for providing registry instances in a Django app's
    :file:`__init__.py` file without needing to import other modules that
    might interfere with the Django initialization process. When accessed
    for the first time, the registry will be imported and constructed.

    This can also speed up the startup process, depending on the complexity
    of registries.

    Args:
        module_path (str):
            The import path of the module containing the registry.

        registry_name (str):
            The class name of the registry.

        **kwargs (dict):
            Keyword arguments to pass to the registry's constructor.

    Returns:
        django.utils.functional.SimpleLazyObject:
        A wrapper that will dynamically load and forward on to the registry.
    """
    def _create_registry():
        mod = import_module(module_path)

        return getattr(mod, registry_name)(**kwargs)

    return SimpleLazyObject(_create_registry)



"""Unit tests for djblets.util.serializers."""

from datetime import datetime

from djblets.testing.testcases import TestCase
from djblets.util.serializers import DjbletsJSONEncoder


class DjbletsJSONEncoderTests(TestCase):
    """Unit tests for djblets.util.serializers.DjbletsJSONEncoder."""

    def test_object_to_json(self):
        """Testing DjbletsJSONEncoder.encode for an object with a to_json()
        method
        """
        class TestObject(object):
            def to_json(self):
                return {
                    'foo': 1,
                }

        obj = TestObject()
        encoder = DjbletsJSONEncoder()

        self.assertEqual(encoder.encode(obj), '{"foo": 1}')

    def test_datetime(self):
        """Testing DjbletsJSONENcoder.encode with datetimes"""
        encoder = DjbletsJSONEncoder()
        self.assertEqual(
            encoder.encode(datetime(2016, 8, 26, 3, 3, 26, 123456)),
            '"2016-08-26T03:03:26"')

    def test_datetime_with_strip_ms(self):
        """Testing DjbletsJSONENcoder.encode with datetimes when using
        strip_datetime_ms=False
        """
        encoder = DjbletsJSONEncoder(strip_datetime_ms=False)
        self.assertEqual(
            encoder.encode(datetime(2016, 8, 26, 3, 3, 26, 123456)),
            '"2016-08-26T03:03:26.123"')

    def test_with_set(self):
        """Testing DjbletsJSONEncoder.encode with set"""
        encoder = DjbletsJSONEncoder()
        self.assertEqual(encoder.encode({1, 2, 9, 4, 4, 10}),
                         '[1, 2, 4, 9, 10]')



"""Filesystem-related utilities."""

import os
import sys


def is_exe_in_path(name):
    """Checks whether an executable is in the user's search path.

    This expects a name without any system-specific executable extension.
    It will append the proper extension as necessary. For example,
    use "myapp" and not "myapp.exe".

    This will return True if the app is in the path, or False otherwise.
    """

    if sys.platform == 'win32' and not name.endswith('.exe'):
        name += ".exe"

    for dir in os.environ['PATH'].split(os.pathsep):
        if os.path.exists(os.path.join(dir, name)):
            return True

    return False



"""Unit tests for djblets.util.templatetags.djblets_deco."""

from django.template import Context, Template, TemplateSyntaxError

from djblets.testing.testcases import TagTest


class BoxTagTests(TagTest):
    """Unit tests for the {% box %} template tag."""

    def test_plain(self):
        """Testing {% box %}"""
        t = Template('{% load djblets_deco %}'
                     '{% box %}content{% endbox %}')

        self.assertHTMLEqual(
            t.render(Context({})),
            '<div class="box-container"><div class="box">'
            '<div class="box-inner">\ncontent\n  '
            '</div></div></div>')

    def test_with_class(self):
        """Testing {% box %} with CSS class"""
        t = Template('{% load djblets_deco %}'
                     '{% box "class" %}content{% endbox %}')

        self.assertHTMLEqual(
            t.render(Context({})),
            '<div class="box-container"><div class="box class">'
            '<div class="box-inner">\ncontent\n  '
            '</div></div></div>')

    def test_with_extra_arg_error(self):
        """Testing {% box %} with extra argument error"""
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load djblets_deco %}'
                     '{% box "class" "foo" %}content{% endbox %}')


class ErrorBoxTests(TagTest):
    """Unit tests for the {% errorbox %} template tag."""

    def test_plain(self):
        """Testing {% errorbox %}"""
        t = Template('{% load djblets_deco %}'
                     '{% errorbox %}content{% enderrorbox %}')

        self.assertHTMLEqual(
            t.render(Context({})),
            '<div class="errorbox">\ncontent\n</div>')

    def test_with_id(self):
        """Testing {% errorbox %} with element ID"""
        t = Template('{% load djblets_deco %}'
                     '{% errorbox "id" %}content{% enderrorbox %}')

        self.assertHTMLEqual(
            t.render(Context({})),
            '<div class="errorbox" id="id">\ncontent\n</div>')

    def test_with_extra_arg_error(self):
        """Testing {% errorbox %} with extra argument error"""
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load djblets_deco %}'
                     '{% box "class" "foo" "foo" %}content{% endbox %}')



"""Decorators to help with API test suites."""

from djblets.util.decorators import simple_decorator


@simple_decorator
def webapi_test_template(test_func):
    """Mark a test function as a template for tests.

    This adds a flag to the test function hinting that it should be processed
    differently. :py:class:`WebAPITestCaseMixin` will replace the docstring to
    match that of the active test suite.
    """
    def _call(*args, **kwargs):
        return test_func(*args, **kwargs)

    _call.is_test_template = True

    return _call



"""Extension hooks for consent requirements."""

from djblets.extensions.hooks import BaseRegistryHook, ExtensionHookPoint
from djblets.privacy.consent.registry import get_consent_requirements_registry


class ConsentRequirementHook(BaseRegistryHook, metaclass=ExtensionHookPoint):
    """Registers a ConsentRequirement for use of personal data."""

    @property
    def registry(self):
        """The registry that the hook interfaces with."""
        return get_consent_requirements_registry()



"""Unit tests for djblets.configforms.views.ConfigPagesView."""

from django.contrib.auth.models import User
from django.http import Http404
from django.test.client import RequestFactory

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class MyTestForm1(ConfigPageForm):
    form_id = 'my-form-1'
    form_title = 'Form 1'

    def save(self):
        pass


class MyTestForm2(ConfigPageForm):
    form_id = 'my-form-2'
    form_title = 'Form 2'


class MyTestForm3(ConfigPageForm):
    form_id = 'my-form-3'
    form_title = 'Form 3'

    def is_visible(self):
        return False


class MyTestPage1(ConfigPage):
    page_id = 'my-page-1'
    form_classes = [MyTestForm1]


class MyTestPage2(ConfigPage):
    page_id = 'my-page-2'
    form_classes = [MyTestForm2]


class MyTestPage3(ConfigPage):
    page_id = 'my-page-3'
    form_classes = [MyTestForm3]


class MyConfigPagesView(ConfigPagesView):
    title = 'My Page Title'
    nav_title = 'My Nav Entry'
    page_classes = [MyTestPage1, MyTestPage2, MyTestPage3]

    css_bundle_names = ['my-css-bundle']
    js_bundle_names = ['my-js-bundle']

    js_model_class = 'MyModel'
    js_view_class = 'MyView'

    def get_js_model_data(self):
        return {
            'my-attr': 'value',
        }

    def get_js_view_data(self):
        return {
            'my-option': 'value',
        }


class ConfigPagesViewTests(TestCase):
    """Unit tests for djblets.configforms.views.ConfigPagesView."""

    def test_dispatch_initial_state(self):
        """Testing ConfigPagesView.dispatch initial state"""
        request = RequestFactory().request()
        request.user = User.objects.create(username='test-user')

        view = MyConfigPagesView()
        view.request = request

        response = view.dispatch(view.request)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(view.pages), 2)
        self.assertIsInstance(view.pages[0], MyTestPage1)
        self.assertIsInstance(view.pages[1], MyTestPage2)
        self.assertEqual(set(view.forms.keys()),
                         {'my-form-1', 'my-form-2'})

    def test_get_context_data(self):
        """Testing ConfigPagesView.get_context_data"""
        request = RequestFactory().request()
        request.user = User.objects.create(username='test-user')

        view = MyConfigPagesView()
        view.request = request

        view.dispatch(view.request)

        self.assertEqual(
            view.get_context_data(),
            {
                'base_template_name': 'base.html',
                'page_title': 'My Page Title',
                'nav_title': 'My Nav Entry',
                'pages_id': 'config_pages',
                'pages': view.pages,
                'css_bundle_names': ['my-css-bundle'],
                'js_bundle_names': ['my-js-bundle'],
                'js_model_class': 'MyModel',
                'js_view_class': 'MyView',
                'js_model_data': {
                    'my-attr': 'value',
                },
                'js_view_data': {
                    'my-option': 'value',
                },
                'forms': list(view.forms.values()),
                'render_sidebar': True,
            })

    def test_post_without_form_target(self):
        """Testing ConfigPagesView.dispatch with POST and no form_target"""
        request = RequestFactory().post('/config/')
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 400)

    def test_post_with_invalid_form_target(self):
        """Testing ConfigPagesView.dispatch with POST and invalid form_target
        """
        request = RequestFactory().post('/config/', {
            'form_target': 'bad',
        })
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()

        with self.assertRaises(Http404):
            view(request)

    def test_post_with_success(self):
        """Testing ConfigPagesView.dispatch with POST and success"""
        request = RequestFactory().post('/config/', {
            'form_target': 'my-form-1',
        })
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/config/')



"""Unit tests for djblets.util.templatetags.djblets_utils."""

import warnings
from collections import OrderedDict
from datetime import datetime, timedelta

from django.http import HttpRequest, QueryDict
from django.template import Context, Template

from djblets.testing.testcases import TagTest, TestCase
from djblets.util.templatetags.djblets_utils import (ageid, escapespaces,
                                                     humanize_list, indent)


class AgeIdTagTests(TagTest):
    """Unit tests for the {% ageid %} template tag."""

    def setUp(self):
        super(AgeIdTagTests, self).setUp()

        self.now = datetime.utcnow()

        self.context = {
            'now': self.now,
            'minus1': self.now - timedelta(1),
            'minus2': self.now - timedelta(2),
            'minus3': self.now - timedelta(3),
            'minus4': self.now - timedelta(4),
        }

    def test_with_now(self):
        """Testing {% ageid %} with now"""
        self.assertEqual(ageid(self.now), 'age1')

    def test_with_now_minus_1_day(self):
        """Testing {% ageid %} with yesterday"""
        self.assertEqual(ageid(self.now - timedelta(1)), 'age2')

    def test_with_now_minus_2_days(self):
        """Testing {% ageid %} with two days ago"""
        self.assertEqual(ageid(self.now - timedelta(2)), 'age3')

    def test_with_now_minus_3_days(self):
        """Testing {% ageid %} with three days ago"""
        self.assertEqual(ageid(self.now - timedelta(3)), 'age4')

    def test_with_now_minus_4_days(self):
        """Testing {% ageid %} with four days ago"""
        self.assertEqual(ageid(self.now - timedelta(4)), 'age5')

    def test_with_non_datetime(self):
        """Testing {% ageid %} with non-datetime object"""
        class Foo:
            def __init__(self, now):
                self.day = now.day
                self.month = now.month
                self.year = now.year

        self.assertEqual(ageid(Foo(self.now)), 'age1')


class AttrTagTests(TestCase):
    """Unit tests for the {% attr %} template tag."""

    def test_with_value(self):
        """Testing {% attr %} with value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "class" %}\n'
                     '{%  if some_bool %}truthy{% endif %}\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context({
                'some_bool': True,
            })),
            '<span class="truthy">')

    def test_without_value(self):
        """Testing {% attr %} with no value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "class" %}\n'
                     '{%  if some_bool %}falsy{% endif %}\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context({
                'some_bool': False,
            })),
            '<span>')

    def test_escapes_value(self):
        """Testing {% attr %} escapes value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" %}<hello>{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="&lt;hello&gt;">')

    def test_condenses_whitespace(self):
        """Testing {% attr %} condenses/strips extra whitespace by default"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" %}\n'
                     'some    \n\n'
                     'value\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="some value">')

    def test_with_nocondense_preserves_whitespace(self):
        """Testing {% attr %} with "nocondense" option preserves whitespace"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" nocondense %}\n'
                     'some    \n\n'
                     'value\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="\nsome    \n\nvalue\n">')


class DefineVarTagTests(TestCase):
    """Unit tests for the {% definevar %} template tag."""

    def test_basic_usage(self):
        """Testing {% definevar %}"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" %}\n'
                     'test{{num}}\n'
                     '{% enddefinevar %}'
                     '{{myvar}}')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '\ntest123\n')

    def test_with_global(self):
        """Testing {% definevar %} with global option"""
        t = Template(
            '{% load djblets_utils %}'
            '{% block main %}'
            '{%  block inner %}'
            '{%   definevar "myvar" global %}{{num}}{% enddefinevar %}'
            '{%  endblock %}'
            '{% endblock %}'
            '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[123]')

    def test_with_strip(self):
        """Testing {% definevar %} with strip option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" strip %}\n'
                     '<span>\n'
                     ' <strong>\n'
                     '  test{{num}}\n'
                     ' </strong>\n'
                     '</span>\n'
                     '{% enddefinevar %}'
                     '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[<span>\n <strong>\n  test123\n </strong>\n</span>]')

    def test_with_spaceless(self):
        """Testing {% definevar %} with spaceless option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" spaceless %}\n'
                     '<span>\n'
                     ' <strong>\n'
                     '  test{{num}}\n'
                     ' </strong>\n'
                     '</span>\n'
                     '{% enddefinevar %}'
                     '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[<span><strong>\n  test123\n </strong></span>]')

    def test_with_unsafe(self):
        """Testing {% definevar %} with unsafe option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" unsafe %}<hello>{% enddefinevar %}'
                     '{{myvar}}')

        self.assertEqual(t.render(Context()), '&lt;hello&gt;')


class EscapeSpacesFilterTests(TestCase):
    """Unit tests for the {{...|escapespaces}} template filter."""

    def test_with_single_space(self):
        """Testing {{...|escapespaces}} with single space"""
        self.assertEqual(escapespaces('Hi there'),
                         'Hi there')

    def test_with_multiple_spaces(self):
        """Testing {{...|escapespaces}} with multiple consecutive spaces"""
        self.assertEqual(escapespaces('Hi  there'),
                         'Hi&nbsp; there')

    def test_with_newline(self):
        """Testing {{...|escapespaces}} with newline"""
        self.assertEqual(escapespaces('Hi  there\n'),
                         'Hi&nbsp; there<br />')


class GetAttrFilterTests(TestCase):
    """Unit tests for the {{...|getattr:...}} template filter."""

    def test_with_attr_found(self):
        """Testing {{...|getattr:...}} with attribute found"""
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"key"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'value')

    def test_with_attr_not_found(self):
        """Testing {{...|getattr:...}} with attribute not found"""
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"bad_key"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'None')

    def test_with_attr_not_found_and_default_if_none(self):
        """Testing {{...|getattr:...}} with attribute not found and
        default_if_none fallback filter
        """
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"bad_key"|default_if_none:"doh"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'doh')


class HumanizeListFilterTests(TestCase):
    """Unit tests for the {{...|humanize_list}} template filter."""

    def test_with_empty_list(self):
        """Testing {{...|humanize_list}} with empty list"""
        self.assertEqual(humanize_list([]),
                         '')

    def test_with_1_item(self):
        """Testing {{...|humanize_list}} with 1 item"""
        self.assertEqual(humanize_list(['a']),
                         'a')

    def test_with_2_items(self):
        """Testing {{...|humanize_list}} with 2 items"""
        self.assertEqual(humanize_list(['a', 'b']),
                         'a and b')

    def test_with_3_items(self):
        """Testing {{...|humanize_list}} with 3 items"""
        self.assertEqual(humanize_list(['a', 'b', 'c']),
                         'a, b and c')

    def test_with_4_items(self):
        """Testing {{...|humanize_list}} with 4 items"""
        self.assertEqual(humanize_list(['a', 'b', 'c', 'd']),
                         'a, b, c, and d')


class IncludeAsStringTagTests(TestCase):
    """Unit tests for the {% include_as_string %} template tag."""

    def test_basic_usage(self):
        """Testing {% include_as_string %}"""
        t = Template('{% load djblets_utils %}'
                     '{% include_as_string template_name %}')

        self.assertEqual(
            t.render(Context({
                'template_name': 'testing/foo.html',
                'foo': 1,
                'bar': 2,
            })),
            "'1 2\\\n'")


class IndentFilterTests(TestCase):
    """Unit tests for the {{...|indent}} template filter."""

    def test_with_default_indent(self):
        """Testing {{...|indent}} with default indentation level"""
        self.assertEqual(indent('foo'), '    foo')

    def test_with_custom_indent(self):
        """Testing {{...|indent}} with custom indentation level"""
        self.assertEqual(indent('foo', 3), '   foo')

    def test_with_multiple_lines(self):
        """Testing {{...|indent}} with multiple lines"""
        self.assertEqual(indent('foo\nbar'),
                         '    foo\n    bar')


class QuerystringTagTests(TestCase):
    """Unit tests for the {% querystring %} template tag."""

    def test_update_basic_usage(self):
        """Testing {% querystring "update" %} basic usage"""
        self.assertEqual(
            self._render_tag(tag='{% querystring "update" "foo=bar" %}',
                             query_str='foo=bar'),
            '?foo=bar')

    def test_update_with_tag_existing_query(self):
        """Testing {% querystring "update" %} with an existing query"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" %}',
            query_str='a=1&b=2')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&b=2&foo=bar'))

    def test_update_with_existing_query_override(self):
        """Testing {% querystring "update" %} with an existing query that gets
        overridden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" %}',
            query_str='foo=foo&bar=baz')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar&bar=baz'))

    def test_update_with_existing_query_with_two_args_override(self):
        """Testing {% querystring "update" %} with two args that get
        overridden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" "qux=baz" %}',
            query_str='foo=foo&bar=bar&baz=baz&qux=qux')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar&bar=bar&baz=baz&qux=baz'))

    def test_update_with_no_value(self):
        """Testing {% querystring "update" %} with no value"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo='))

    def test_update_with_multiple_values(self):
        """Testing {% querystring "update" %} with multiple values"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar=baz" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar=baz'))

    def test_update_with_empty_value(self):
        """Testing {% querystring "update" %} with empty value"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo='))

    def test_update_with_no_key(self):
        """Testing {% querystring "update" %} with no key"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "=foo" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('=foo'))

    def test_update_with_querystring_key_overide(self):
        """Testing {% querystring "update" %} by updating multiple values of a
        key value gets overriden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "a=1" "a=2" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&a=2'))

    def test_with_updating_multiple_values_of_a_key(self):
        """Testing {% querystring "update" %} by updating multiple values of a
        key value
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "a=1&a=2" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&a=1&a=2'))

    def test_append_with_basic_usage(self):
        """Testing {% querystring "append" %} with appending on to an existing
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "foo=baz" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&foo=baz&bar=bar'))

    def test_append_with_multiple_values_and_same_key(self):
        """Testing {% querystring "append" %} with appending multiple values of
        a key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "a=1&a=2&a=3" %}',
            query_str='a=0&&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=0&a=1&a=2&a=3&b=2&c=3'))

    def test_append_with_multiple_values_and_same_key_seperated(self):
        """Testing {% querystring "append" %} with appending multiple values of
        a key fragment
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "a=1" "a=2" "a=3" %}',
            query_str='a=0&&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=0&a=1&a=2&a=3&b=2&c=3'))

    def test_append_with_new_key(self):
        """Testing {% querystring "append" %} with appending new key-value
        pair
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "d=4" %}',
            query_str='a=1&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&b=2&c=3&d=4'))

    def test_remove_with_basic_usage(self):
        """Testing {% querystring "remove" %} by removing a single instance of
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('bar=bar'))

    def test_remove_with_key_not_in_querystring(self):
        """Testing {% querystring "remove" %} by attempting to remove a
        non-existing key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "baz" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&bar=bar'))

    def test_remove_with_key_appearing_multiple_times(self):
        """Testing {% querystring "remove" %} by removing all instances of a
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" %}',
            query_str='foo=foo&foo=bar&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('bar=bar'))

    def test_remove_for_specific_key_value_pairs(self):
        """Testing {% querystring "remove" %} by removing a specific key-value
        pair
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "a=4" %}',
            query_str='a=1&a=2&a=3&a=4')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&a=2&a=3&'))

    def test_remove_with_no_key(self):
        """Testing {% querystring "remove" %} by removing a value with
        no key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "=foo" %}',
            query_str='foo=foo&foo=bar&baz=baz&=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&foo=bar&baz=baz'))

    def test_remove_with_no_value(self):
        """Testing {% querystring "remove" %} by removing a key with no
        value
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo=" %}',
            query_str='foo=foo&foo=bar&foo=&baz=baz')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('baz=baz'))

    def test_remove_with_multiple_removes(self):
        """Testing {% querystring "remove" %} by removing multiple keys and
        values
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" "bar" "baz=1" %}',
            query_str='foo=foo&bar=bar&foo=&baz=1&qux=qux')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('qux=qux'))

    def test_remove_with_multiple_specific_values(self):
        """Testing {% querystring "remove" %} by removing multiple specific
        key-value pairs
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo=1" "foo=2" %}',
            query_str='foo=1&foo=2&foo=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo=3'))

    def _render_tag(self, tag, query_str):
        """Returned a rendered template tag using a query string.

        This will render a ``querystring`` template using the provided template
        tag, with auto-escaping turned off, and with the given query string as
        would be provided in a URL.

        Args:
            tag (unicode):
                The template tag line (``{% querystring ... %}``).

            query_str (unicode):
                The query string to simulate being in the URL.

        Returns:
            unicode:
            The rendered content, without escaping.
        """
        t = Template('{%% load djblets_utils %%}'
                     '{%% autoescape off %%}%s{%% endautoescape %%}'
                     % tag)

        request = HttpRequest()

        if query_str:
            request.GET = QueryDict(query_str)

        return t.render(Context({
            'request': request,
        })).replace('&amp;', '&')



#!/usr/bin/env python3

from __future__ import unicode_literals

import os
import sys

scripts_dir = os.path.abspath(os.path.dirname(__file__))

# Source root directory
sys.path.insert(0, os.path.abspath(os.path.join(scripts_dir, '..', '..')))

# Script config directory
sys.path.insert(0, os.path.join(scripts_dir, 'conf'))

from djblets.dependencies import django_version

import __main__
__main__.__requires__ = ['Django' + django_version]

# This is required to re-import.
import pkg_resources

import django
from django.core.management import call_command


if __name__ == '__main__':
    os.putenv('FORCE_BUILD_MEDIA', '1')
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'djblets.settings')

    if hasattr(django, 'setup'):
        # Django >= 1.7
        django.setup()

    # This will raise a CommandError or call sys.exit(1) on failure.
    call_command('collectstatic', interactive=False, verbosity=2)



"""Feature-related decorators."""

from functools import wraps

from django.http import HttpResponseNotFound


def _404(*args, **kwargs):
    return HttpResponseNotFound()


def feature_required(feature, not_enabled_view=None):
    """Require a feature to execute a view.

    If the feature is not enabled, the ``not_enabled_view`` will be called
    instead.

    Args:
        feature (djblets.features.feature.Feature):
            The feature that must be enabled.

        not_enabled_view (callable):
            The view that will be called when the feature is not enabled. This
            defaults to a view that returns an :http:`404`.

    Returns:
        callable:
        The decorator.
    """
    if not_enabled_view is None:
        not_enabled_view = _404

    def decorator(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            if not feature.is_enabled():
                return not_enabled_view(*args, **kwargs)

            return f(*args, **kwargs)

        return decorated

    return decorator



"""Siteconfig definitions for the log app."""

from djblets.log import DEFAULT_LOG_LEVEL

settings_map = {
    'logging_enabled':         'LOGGING_ENABLED',
    'logging_directory':       'LOGGING_DIRECTORY',
    'logging_allow_profiling': 'LOGGING_ALLOW_PROFILING',
    'logging_level':           'LOGGING_LEVEL',
}

defaults = {
    'logging_enabled':         False,
    'logging_directory':       None,
    'logging_allow_profiling': False,
    'logging_level':           DEFAULT_LOG_LEVEL,
}



"""Miscellaneous utilities."""

import logging
import os
import signal
import sys
from contextlib import contextmanager

from django.utils.translation import gettext as _


logger = logging.getLogger(__name__)


def kill_process(pid):
    """Kill a process."""
    # This is necessary because we need to continue supporting Python 2.5,
    # which doesn't have Popen.kill(). This is inspired by
    # http://stackoverflow.com/questions/1064335
    if sys.platform == 'win32':
        import ctypes
        PROCESS_TERMINATE = 1
        handle = ctypes.windll.kernel32.OpenProcess(
            PROCESS_TERMINATE, False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, -1)
        ctypes.windll.kernel32.CloseHandle(handle)
    else:
        os.kill(pid, signal.SIGKILL)


@contextmanager
def controlled_subprocess(process_name, process):
    """
    A context manager for a subprocess that guarantees that a process
    is terminated, even if exceptions are thrown while using it.

    The process_name argument is used for logging when the process goes
    down fighting.  The process argument is a process returned by
    subprocess.Popen.

    Example usage:

    process = subprocess.Popen(['patch', '-o', newfile, oldfile])

    with controlled_subprocess("patch", process) as p:
        # ... do things with the process p

    Once outside the with block, you can rest assured that the subprocess
    is no longer running.
    """

    caught_exception = None

    try:
        yield process
    except Exception as e:
        caught_exception = e

    # If we haven't gotten a returncode at this point, we assume the
    # process is blocked.  Let's kill it.
    if process.returncode is None and process.poll() is None:
        logger.warning(
            _("The process '%(name)s' with PID '%(pid)s' did not exit "
              "cleanly and will be killed automatically.")
            % {
                'name': process_name,
                'pid': process.pid,
            })

        kill_process(process.pid)

        # Now that we've killed the process, we'll grab the return code,
        # in order to clear the zombie.
        process.wait()

    # If we caught an exception earlier, re-raise it.
    if caught_exception:
        raise caught_exception



"""Helpers for unit tests working with features."""

from contextlib import contextmanager

from djblets.features.feature import Feature
from djblets.features.registry import get_features_registry


@contextmanager
def override_feature_checks(feature_states):
    """Override multiple features for a test.

    Unit tests can make use of this context manager to ensure that one or more
    features have particular enabled/disabled states before executing any code
    dependent on those features.

    Only the provided features will be modified, with all other feature logic
    falling back to the default behavior for the configured feature checker.

    Version Change:
        1.0.13:
        ``feature_states`` now accepts a
        :py:class:`~djblets.features.feature.Feature` instance as a key.

    Args:
        feature_states (dict):
            A dictionary of feature IDs or instances to booleans (representing
            whether the feature is enabled).

    Example:
        .. code-block:: python

           from myproject.features import my_feature_3

           feature_states = {
               'my-feature-1': True,
               'my-feature-2': False,
               my_feature_3: True,
           }

           with override_feature_checks(feature_states):
               # Your test code here.
    """
    registry = get_features_registry()
    old_state = []

    for feature, enabled in feature_states.items():
        if not isinstance(feature, Feature):
            feature = registry.get_feature(feature)

        old_state.append({
            'feature': feature,
            'func': feature.is_enabled,
        })

        feature.is_enabled = \
            lambda _is_enabled=enabled, **kwargs: _is_enabled

    try:
        yield
    finally:
        for feature_info in old_state:
            feature_info['feature'].is_enabled = feature_info['func']


@contextmanager
def override_feature_check(feature_id, enabled):
    """Override a feature for a test.

    Unit tests can make use of this context manager to ensure that a specific
    feature has a particular enabled/disabled state before executing any code
    dependent on that feature.

    Only the provided feature will be modified, with all other feature logic
    falling back to the default behavior for the configured feature checker.

    Args:
        feature_id (unicode or djblets.features.feature.Feature):
            The ID or instance of the feature to override.

        enabled (bool):
            The enabled state for the feature.

    Example:
        .. code-block:: python

           from myproject.features import my_feature_2

           with override_feature_check('my-feature', enabled=False):
               # Your test code here.

           with override_feature_check(my_feature_2, enabled=True):
               # Your test code here.
    """
    with override_feature_checks({feature_id: enabled}):
        yield



"""Compatibility for features removed in Python 3.

Note that this module is not considered API-stable. Changes may be made that
remove or alter functionality in the future. Please use at your own risk.
"""

cmp = lambda x, y: (x > y) - (x < y)


__all__ = (
    'cmp',
)
